import operator
import functools
import itertools
import time
import re

import Bio.SeqUtils
import pysam

import importlib

from handygenome.interval import Interval
import handygenome.deco as deco
import handygenome.variant.infoformat as infoformat
import handygenome.variant.varianthandler as varianthandler
import handygenome.annotation.annotitem as annotitem
import handygenome.annotation.ensembl_parser as ensembl_parser
import handygenome.annotation.ensembl_rest as ensembl_rest
import handygenome.annotation.veplib as veplib
import handygenome.annotation.customfile as customfile
import handygenome.annotation.rnalib as rnalib
import handygenome.variant.vcfspec as libvcfspec


META_ID_PREFIX = {
    "transcript": "transcript",
    "regulatory": "regulatory",
    "motif": "motif",
    "repeat": "repeat",
}
META_ID_SUFFIX = {"plain": "", "bnd1": "_bnd1", "bnd2": "_bnd2"}
META_NUMS = {"transcript": ".", "regulatory": ".", "motif": ".", "repeat": "."}
META_DESCRIPTIONS = {
    "transcript": "Transcript feature annotations encoded as a string",
    "regulatory": "Regulatory element feature annotations encoded as a string",
    "motif": "Transcription factor binding motif feature annotations encoded as a string",
    "repeat": "Repeat feature annotations encoded as a string",
}
META_DESCRIPTIONS_SUFFIXES = {
    "plain": "for non-breakend variants (e.g. snv, indel, cnv)",
    "bnd1": "for breakend 1",
    "bnd2": "for breakend 2",
}

HGVSP_PAT = re.compile('^(.+):p.([^0-9]+)([0-9]+)([^0-9]+)$')



###############################################


#REGULATORY_FEATURE_TYPES = {
#    "regulatory": {
#        "Promoter": "promoter",
#        "Promoter Flanking Region": "promoter_flank",
#        "CTCF Binding Site": "CTCFBS",
#        "TF binding site": "TFBS",
#        "Enhancer": "enhancer",
#        "Open chromatin": "open_chromatin",
#    },
#    "overlap": {
#        "Predicted promoter": "promoter",
#        "Predicted promoter flanking region": "promoter_flank",
#        "CTCF binding site": "CTCFBS",
#        "Transcription factor binding site": "TFBS",
#        "Predicted enhancer region": "enhancer",
#        "Open chromatin region": "open_chromatin",
#    },
#    "vep": {
#        "promoter": "promoter",
#        "promoter_flanking_region": "promoter_flank",
#        "CTCF_binding_site": "CTCFBS",
#        "TF_binding_site": "TFBS",
#        "enhancer": "enhancer",
#        "open_chromatin_region": "open_chromatin",
#    },
#}

BIOTYPE_REGULATORY_CLASSES = {
    "promoter": (
        "Promoter",
        "Predicted promoter",
        "promoter",
    ),
    "promoter_flank": (
        "Promoter Flanking Region",
        "Predicted promoter flanking region",
        "promoter_flanking_region",
    ),
    "CTCFBS": (
        "CTCF Binding Site",
        "CTCF binding site",
        "CTCF_binding_site",
    ),
    "TFBS": (
        "TF binding site",
        "Transcription factor binding site",
        "TF_binding_site",
    ),
    "enhancer": (
        "Enhancer",
        "Predicted enhancer region",
        "enhancer",
    ),
    "open_chromatin": (
        "Open chromatin",
        "Open chromatin region",
        "open_chromatin_region",
    ),
}
BIOTYPE_REGULATORY_VALUES = set(
    itertools.chain.from_iterable(BIOTYPE_REGULATORY_CLASSES.values())
)


# reference: https://asia.ensembl.org/info/genome/variation/prediction/predicted_data.html
CONSEQUENCE_TRANSCRIPT_CLASSES = {
    "splice_region_involved": (
        "splice_acceptor_variant",
        "splice_donor_variant",
        "splice_donor_region_variant",  # added on 220422
        "splice_region_variant",
        "splice_polypyrimidine_tract_variant",  # added on 220329
        "splice_donor_5th_base_variant",  # added on 220401
    ),
    "splice_acceptor_involved": ("splice_acceptor_variant",),
    "splice_donor_involved": (
        "splice_donor_variant",
        "splice_donor_region_variant",
        "splice_donor_5th_base_variant",  # added on 220401
    ),
    "5pUTR_involved": ("5_prime_UTR_variant",),
    "3pUTR_involved": ("3_prime_UTR_variant",),
    "protein_altering": (
        "stop_gained",
        "frameshift_variant",
        "stop_lost",
        "start_lost",
        "inframe_insertion",
        "inframe_deletion",
        "missense_variant",
        "protein_altering_variant",
    ),
    "not_protein_altering": (
        "splice_acceptor_variant",
        "splice_donor_variant",
        "splice_region_variant",
        "start_retained_variant",
        "stop_retained_variant",
        "synonymous_variant",
        "mature_miRNA_variant",
        "5_prime_UTR_variant",
        "3_prime_UTR_variant",
        "non_coding_transcript_exon_variant",
        "intron_variant",
        "NMD_transcript_variant",
        "non_coding_transcript_variant",
        "upstream_gene_variant",
        "downstream_gene_variant",
        "TF_binding_site_variant",
        "regulatory_region_variant",
        "intergenic_variant",
    ),
    "synonymous": ("synonymous_variant",),
    "missense": ("missense_variant",),
    "frameshift": ("frameshift_variant",),
    "inframe": ("inframe_deletion",),
    "stopgain": ("stop_gained",),
    "stoplost": ("stop_lost",),
    "startlost": ("start_lost",),
    "unclassified": (
        "incomplete_terminal_codon_variant",
        "coding_sequence_variant",
    ),
    "SV_consequence": (
        "transcript_ablation",
        "transcript_amplification",
        "TFBS_ablation",
        "TFBS_amplification",
        "regulatory_region_ablation",
        "regulatory_region_amplification",
        "feature_elongation",
        "feature_truncation",
    ),
}


CONSEQUENCE_TRANSCRIPT_VALUES = set(
    itertools.chain.from_iterable(CONSEQUENCE_TRANSCRIPT_CLASSES.values())
)


BIOTYPE_TRANSCRIPT_CLASSES = {
    # coding
    "coding": (
        "IG_C_gene",
        "IG_D_gene",
        "IG_J_gene",
        "IG_V_gene",
        "ccds_gene",
        "mRNA",
        "protein_coding",
    ),
    "IG": (
        "IG_C_gene",
        "IG_D_gene",
        "IG_J_gene",
        "IG_V_gene",
    ),
    "TR": (
        "TR_C_gene",
        "TR_D_gene",
        "TR_J_gene",
        "TR_V_gene",
    ),
    # noncoding
    "noncoding": (
        "3prime_overlapping_ncRNA",  # updated on 220421; ENST00000606379
        #"3prime_overlapping_ncrna",
        "Mt_rRNA",
        "Mt_tRNA",
        "RNase_MRP_RNA",
        "RNase_P_RNA",
        "Y_RNA",
        "antisense_RNA",
        "antisense",  # updated on 220316; ENST00000442411
        "lncRNA",
        "lincRNA",
        "miRNA",
        "misc_RNA",
        "non_stop_decay",
        "nonsense_mediated_decay",
        "processed_transcript",
        "rRNA",
        "retained_intron",
        "ribozyme",
        "sRNA",
        "scRNA",
        "scaRNA",
        "snRNA",
        "snoRNA",
        "tRNA",
        "telomerase_RNA",
        "vault_RNA",
        "sense_overlapping",
        "sense_intronic",  # updated on 220421; ENST00000428191
    ),
    "rRNA": (
        "Mt_rRNA",
        "rRNA",
    ),
    "tRNA": (
        "Mt_tRNA",
        "tRNA",
    ),
    "miRNA": ("miRNA",),
    "lncRNA": (
        "lncRNA",
        "lincRNA",
    ),
    "NMD": ("nonsense_mediated_decay",),
    "NSD": ("non_stop_decay",),
    "antisense": (
        "antisense_RNA",
        "antisense",
    ),
    # pseudogenes
    "pseudogene": (
        "IG_C_pseudogene",
        "IG_J_pseudogene",
        "IG_V_pseudogene",
        "IG_pseudogene",
        "TR_J_pseudogene",
        "TR_V_pseudogene",
        "ncRNA_pseudogene",
        "polymorphic_pseudogene",
        "processed_pseudogene",
        "pseudogene",
        "rRNA_pseudogene",
        "transcribed_processed_pseudogene",
        "transcribed_pseudogene",
        "transcribed_unitary_pseudogene",
        "transcribed_unprocessed_pseudogene",
        "translated_processed_pseudogene",
        "translated_unprocessed_pseudogene",
        "unitary_pseudogene",
        "unprocessed_pseudogene",
    ),
    "IG_pseudogene": (
        "IG_C_pseudogene",
        "IG_J_pseudogene",
        "IG_V_pseudogene",
        "IG_pseudogene",
    ),
    "TR_pseudogene": (
        "TR_J_pseudogene",
        "TR_V_pseudogene",
    ),
    "processed_pseudogene": (
        "processed_pseudogene",
        "transcribed_processed_pseudogene",
        "transcribed_unprocessed_pseudogene",
        "translated_processed_pseudogene",
        "translated_unprocessed_pseudogene",
        "unprocessed_pseudogene",
    ),
    "unprocessed_pseudogene": (
        "transcribed_unprocessed_pseudogene",
        "translated_unprocessed_pseudogene",
        "unprocessed_pseudogene",
    ),
    "unitary_pseudogene": (
        "transcribed_unitary_pseudogene",
        "unitary_pseudogene",
    ),
    "translated_pseudogene": (
        "translated_processed_pseudogene",
        "translated_unprocessed_pseudogene",
    ),
    "transcribed_pseudogene": (
        "transcribed_processed_pseudogene",
        "transcribed_pseudogene",
        "transcribed_unitary_pseudogene",
        "transcribed_unprocessed_pseudogene",
    ),
    # unknown types
    "unknown_type": (
        "LRG_gene",
        "TEC",
        "aligned_transcript",
        "cdna_update",
        "guide_RNA",
        "other",
    ),
}

BIOTYPE_TRANSCRIPT_VALUES = set(
    itertools.chain.from_iterable(BIOTYPE_TRANSCRIPT_CLASSES.values())
)


################
# Base classes #
################

class EnsemblFeature(annotitem.AnnotItemInfoSingle):
    @property
    def is_overlapping(self):
        return self["distance"] is None


class EnsemblFeatureSetBase(annotitem.AnnotItemInfoSingle):
    #@classmethod
    #def from_vr_base_EnsemblFeatureSet(cls, vr):
    #    result = cls.from_vr_base(vr)
    #    for key in tuple(result.keys()):
    #        result[key] = cls.unit_class.from_dict(result[key])
#
#        return result

    @classmethod
    def from_raw_value(cls, vr_value):
        result = cls.init_nonmissing()
        decoded_raw = cls.decode(vr_value)
        for key, val in decoded_raw.items():
            unit = cls.unit_class.init_nonmissing()
            unit.update(val)
            result[key] = unit
        return result

    @property
    def overlapping(self):
        result = self.__class__()
        for key, val in self.items():
            if val.is_overlapping:
                result[key] = val
        return result

    def get_self_show(self):
        return {key: val.get_self_show() for key, val in self.items()}

    def add_feature(self, feature):
        self[feature["id"]] = feature

    def update_other(self, other, overwrite=True, create_new=True):
        """Args:
            other: another EnsemblFeatureSetBase object
        """
        for ID, feature in other.items():
            if ID in self.keys():
                self[ID].update_dict(feature, overwrite=overwrite)
            else:
                if create_new:
                    self[ID] = feature

    def get_distance_setter_snv(self, vcfspec):
        pos0 = vcfspec.pos0
        def distance_setter(feature_range0):
            if pos0 < feature_range0.start:
                distance = feature_range0.start - pos0
            elif pos0 >= feature_range0.stop:
                distance = feature_range0.stop - 1 - pos0
            else:
                distance = None
            return distance
        return distance_setter

    def get_distance_setter_ins(self, vcfspec):
        pos0 = vcfspec.pos0
        def distance_setter(feature_range0):
            if pos0 <= feature_range0.start - 1:
                distance = feature_range0.start - pos0 - 1
            elif pos0 >= feature_range0.stop - 1:
                distance = feature_range0.stop - 1 - pos0
            else:
                distance = None
            return distance
        return distance_setter

    def get_distance_setter_interval(self, var_range0):
        def distance_setter(feature_range0):
            if var_range0.stop <= feature_range0.start:
                distance = feature_range0.start - var_range0.stop + 1
            elif var_range0.start >= feature_range0.stop:
                distance = feature_range0.stop - 1 - var_range0.start
            else:
                distance = None
            return distance
        return distance_setter

    def get_distance_setter_del(self, vcfspec):
        var_range0 = range(vcfspec.pos, vcfspec.pos + len(vcfspec.ref) - 1)
        return self.get_distance_setter_interval(var_range0)

    def get_distance_setter_bnd(self, pos):
        vcfspec = libvcfspec.Vcfspec(chrom=None, pos=pos, ref=None, alts=None)
        return self.get_distance_setter_snv(vcfspec)

    def set_missing_distances_base(self, distance_setter):
        """Sets distance for features not derived from VEP (e.g. repeat
        elements). Simulates the distance given by VEP.
        """
        for feature in self.values():
            if "distance" not in feature:
                # features derived from "ensembl rest overlap"
                #   do not have "distance" value
                assert "start0" in feature and "end0" in feature, (
                    f'"start0" or "end0" missing for this feature:\n{feature}'
                )
                feature_range0 = range(feature["start0"], feature["end0"])
                feature["distance"] = distance_setter(feature_range0)


class EnsemblFeatureSetPlain(EnsemblFeatureSetBase):
    def get_fetch_interval(self, vcfspec, distance, chromdict):
        return Interval(
            chrom=vcfspec.chrom,
            start1=max(1, vcfspec.pos - distance),
            end1=min(chromdict[vcfspec.chrom], vcfspec.pos + distance),
        )

    def set_missing_distances(self, vcfspec, alt_idx=0):
        mttype = vcfspec.get_mutation_type(alt_idx)
        if mttype == "snv":
            distance_setter = self.get_distance_setter_snv(vcfspec)
        elif mttype == "ins":
            distance_setter = self.get_distance_setter_ins(vcfspec)
        elif mttype in ("del", "delins", "mnv", "cpgmet"):
            distance_setter = self.get_distance_setter_del(vcfspec)
        else:
            raise Exception(f"Unexpected mutation type: {mttype}")

        self.set_missing_distances_base(distance_setter)


class EnsemblFeatureSetBreakend(EnsemblFeatureSetBase):
    pass


##################
# Higher classes #
##################

class Transcript(EnsemblFeature):
    def __getitem__(self, key):
        if key == 'consequence_flags':
            return self.consequence_flags
        elif key == 'subtype_flags':
            return self.subtype_flags
        elif key == 'exon_ranges':
            return self.exon_ranges
        elif key == 'intron_ranges':
            return self.intron_ranges
        else:
            return super().__getitem__(key)

    def get_self_show(self):
        self_show = dict(self)
        for key in (
            'consequence_flags', 
            'subtype_flags',
            'exon_ranges',
            'intron_ranges',
        ):
            try:
                self_show[key] = self[key]
            except KeyError:
                pass

        return self_show

    @functools.cached_property
    def consequence_flags(self):
        consequences = set(self["consequences"])
        if not consequences.issubset(CONSEQUENCE_TRANSCRIPT_VALUES):
            unexpected_consequences = consequences.difference(
                CONSEQUENCE_TRANSCRIPT_VALUES
            )
            raise Exception(
                f'Unexpected consequence value; consequence: {unexpected_consequences}; feature_id: {self["id"]}'
            )
        result = dict()
        for key, val in CONSEQUENCE_TRANSCRIPT_CLASSES.items():
            result[key] = bool(consequences.intersection(val))
        return result

    @functools.cached_property
    def subtype_flags(self):
        if self["biotype"] not in BIOTYPE_TRANSCRIPT_VALUES:
            raise Exception(
                f'Unexpected transcript biotype value; biotype: {self["biotype"]}; feature_id: {self["id"]}'
            )
        result = dict()
        for key, val in BIOTYPE_TRANSCRIPT_CLASSES.items():
            result[key] = self["biotype"] in val
        return result

    @functools.cached_property
    def exon_ranges(self):
        if 'exon_borders' in self.keys():
            if self['exon_borders'] is None:
                return None
            else:
                return [range(*x) for x in self['exon_borders']]
        else:
            return None

    @functools.cached_property
    def intron_ranges(self):
        exon_ranges = self.exon_ranges
        if exon_ranges is None:
            result = None
        else:
            if len(exon_ranges) == 1:
                result = None
            else:
                assert self['is_forward'] is not None
                result = list()
                if self['is_forward']:
                    for idx in range(len(exon_ranges) - 1):
                        result.append(
                            range(exon_ranges[idx].stop, exon_ranges[idx + 1].start)
                        )
                else:
                    for idx in range(len(exon_ranges) - 1):
                        result.append(
                            range(exon_ranges[idx + 1].stop, exon_ranges[idx].start)
                        )
        return result

    def get_hgvsc_genename(self):
        if self['hgvsc'] is None:
            return None
        else:
            #raw_split = self['hgvsc'].split(':n.')
            raw_split = re.split(r':.\.', self['hgvsc'])
            if self['gene_name'] is None:
                return ' '.join(raw_split)
            else:
                return self['gene_name'] + ' ' + raw_split[1]

            #if self['gene_name'] is None:
            #    return self['hgvsc']
            #else: 
            #    return re.sub(r'^(.+)(:.\.)(.+)$', self['gene_name'] + r'\2\3', self['hgvsc'])
                #return self['gene_name'] + ':c.' + self['hgvsc'].split(':c.')[1]

    def get_hgvsp_genename(self, one_letter=False):
        if self['hgvsp'] is None:
            return None
        else:
            raw_split = self['hgvsp'].split(':p.')
            aachange = raw_split[1]
            if one_letter:
                aachange = re.sub(
                    '[A-Z][a-z]{2}', 
                    lambda x: Bio.SeqUtils.seq1(x.group(0)), 
                    aachange,
                )
            return self['gene_name'] + ' ' + aachange


class TranscriptSet(EnsemblFeatureSetPlain):
    unit_class = Transcript

    @property
    def list(self):
        return list(self.values())

    @property
    def canonical(self):
        result = self.__class__.init_nonmissing()
        for key, val in self.items():
            if val['is_canonical']:
                result[key] = val
        return result

    @property
    def canon_ovlp(self):
        result = self.__class__.init_nonmissing()
        for key, val in self.items():
            if (val['is_canonical'] and val.is_overlapping):
                result[key] = val
        return result

    @property
    def is_exonic(self):
        return any(tr['is_exon_involved'] for tr in self.canon_ovlp.values())

    @property
    def is_intronic(self):
        return any(tr['is_intron_involved'] for tr in self.canon_ovlp.values())

    @property
    def is_protein_altering(self):
        return any(tr['consequence_flags']['protein_altering'] for tr in self.canon_ovlp.values())

    def get_gene_names(self, canonical=True, overlap=True):
        if canonical:
            if overlap:
                features = self.canon_ovlp.values()
            else:
                features = self.canonical.values()
        else:
            if overlap:
                features = self.overlapping.values()
            else:
                features = self.values()

        return set(x['gene_name'] for x in features)

    def update_ensembl_gff(self, vcfspec, chromdict, distance, tabixfile_geneset, refver=None, fill_missing_canonical=True):
        assert not ((refver is None) and fill_missing_canonical), (
            f'If "fill_missing_canonical" is True, "refver" must be given.'
        )

        fetch_interval = self.get_fetch_interval(vcfspec, distance, chromdict)
        transcript_set = customfile.fetch_transcript(
            fetch_interval.chrom,
            fetch_interval.start0,
            fetch_interval.end0,
            tabixfile_geneset,
        )

        if fill_missing_canonical:
            for ID, transcript in transcript_set.items():
                if "is_canonical" not in transcript.keys():
                    raw_result = ensembl_rest.lookup_id(
                        ID=transcript["id"], refver=refver, expand=False
                    )
                    time.sleep(0.1)
                        # You have exceeded the limit of 15 requests per second; please reduce your concurrent connections
                    transcript["is_canonical"] = (raw_result["is_canonical"] == 1)

        self.update_other(transcript_set, overwrite=False, create_new=False)

    def update_postvep(self, vcfspec, chromdict, distance, tabixfile_geneset):
        self.update_ensembl_gff(vcfspec, chromdict, distance, tabixfile_geneset, refver=None, fill_missing_canonical=False)


class TranscriptSetALTlist(annotitem.AnnotItemInfoALTlist):
    meta = {
        "ID": "transcripts",
        "Number": "A",
        "Type": "String",
        "Description": "Transcript annotations encoded as a string, one for each ALT allele",
    }

    unit_class = TranscriptSet

    @classmethod
    def from_vr(cls, vr):
        return cls.from_vr_base(vr)
        #result = cls.from_vr_base(vr)
        #for feature_set in result:
        #    for key in tuple(feature_set.keys()):
        #        feature_set[key] = cls.unit_class.unit_class.from_dict(feature_set[key])
        #return result

    def write(self, vr):
        self.write_base(vr)


class Regulatory(EnsemblFeature):
    def __getitem__(self, key):
        if key == 'subtype_flags':
            return self.subtype_flags
        else:
            return super().__getitem__(key)

    def get_self_show(self):
        self_show = dict(self)
        for key in (
            'subtype_flags',
        ):
            try:
                self_show[key] = self[key]
            except KeyError:
                pass

        return self_show

    @functools.cached_property
    def subtype_flags(self):
        if self["biotype"] not in BIOTYPE_REGULATORY_VALUES:
            raise Exception(
                f'Unexpected regulatory biotype value;\n{dict(self)}'
            )
        result = dict()
        for key, val in BIOTYPE_REGULATORY_CLASSES.items():
            result[key] = self["biotype"] in val
        return result


class RegulatorySet(EnsemblFeatureSetPlain):
    meta = {
        "ID": "regulatory_elements",
        "Number": "1",
        "Type": "String",
        "Description": "Regulatory element annotations encoded as a string.",
    }
    unit_class = Regulatory

    @classmethod
    def from_vr(cls, vr):
        return cls.from_vr_base(vr)

    def write(self, vr):
        self.write_base(vr)

    def update_ensembl_gff(self, vcfspec, chromdict, distance, tabixfile_regulatory):
        fetch_interval = self.get_fetch_interval(vcfspec, distance, chromdict)
        regulatory_set = customfile.fetch_regulatory(
            fetch_interval.chrom,
            fetch_interval.start0,
            fetch_interval.end0,
            tabixfile_regulatory,
        )
        self.update_other(regulatory_set, overwrite=False, create_new=True)


class Motif(EnsemblFeature):
    def get_self_show(self):
        return dict(self)


class MotifSet(EnsemblFeatureSetPlain):
    meta = {
        "ID": "TF_binding_motifs",
        "Number": "1",
        "Type": "String",
        "Description": "Transcription factor binding motif annotations encoded as a string.",
    }
    unit_class = Motif

    @classmethod
    def from_vr(cls, vr):
        return cls.from_vr_base(vr)

    def write(self, vr):
        self.write_base(vr)


class Repeat(EnsemblFeature):
    def __getitem__(self, key):
        if key == 'id':
            return self.id
        else:
            return super().__getitem__(key)

    def get_self_show(self):
        self_show = dict(self)
        for key in (
            'id',
        ):
            try:
                self_show[key] = self[key]
            except KeyError:
                pass

        return self_show

    @functools.cached_property
    def id(self):
        return '_'.join(
            [
                self['name'],
                self['chrom'], 
                str(self['start1']), 
                str(self['end1']), 
            ]
        )

class RepeatSet(EnsemblFeatureSetPlain):
    meta = {
        "ID": "repeat_elements",
        "Number": "1",
        "Type": "String",
        "Description": "Repeat sequence annotations encoded as a string.",
    }
    unit_class = Repeat

    @classmethod
    def from_vr(cls, vr):
        return cls.from_vr_base(vr)

    def write(self, vr):
        self.write_base(vr)

    def update_repeatmasker_bed(self, vcfspec, chromdict, distance, tabixfile_repeats):
        fetch_interval = self.get_fetch_interval(vcfspec, distance, chromdict)
        repeat_set = customfile.fetch_repeat(
            fetch_interval.chrom,
            fetch_interval.start0,
            fetch_interval.end0,
            tabixfile_repeats,
        )
        self.update_other(repeat_set, overwrite=False, create_new=True)





import os
import tempfile
import shutil
import subprocess

import pyranges as pr
import pandas as pd

import handygenome.tools as tools
import handygenome.bameditor as bameditor
import handygenome.cnv.misc as cnvmisc


MOSDEPTH_PATH = '/home/users/pjh/tools/miniconda/221104/miniconda3/envs/mosdepth/bin/mosdepth'


def get_mosdepth_args(prefix, bam_path, t, use_median, no_perbase=True, bed_path=None, window_size=None):
    if (
        (bed_path is not None)
        and (window_size is not None)
    ):
        raise Exception(f'"window_size" and "bed_path" arguments must not be used at the same time.')

    args = [MOSDEPTH_PATH, '-x', '-t', str(t)]

    if use_median:
        args.append('-m')

    if no_perbase:
        args.append('-n')

    if bed_path is not None:
        args.extend(('-b', bed_path))
    elif window_size is not None:
        args.extend(('-b', str(window_size)))

    args.extend([prefix, bam_path])

    return args


def load_mosdepth_output(filename, depth_colname='mean_depth', as_gr=True):
    """Only for *.regions.bed.gz file"""

    with tools.openfile(filename) as infile:
        firstline = next(infile)
    firstline_sp = firstline.split('\t')
    if firstline_sp[0] == 'Chromosome':
        if firstline_sp[-1] != depth_colname:
            raise Exception(f'Last column must be "{depth_colname}"')
        header = 0
        names = None
    else:
        header = None
        names = ['Chromosome', 'Start', 'End', depth_colname]
        
    df = pd.read_csv(
        filename, 
        sep='\t', 
        header=header,
        names=names, 
        dtype={'Chromosome': str, 'Start': int, 'End': int, depth_colname: float},
    )
    if as_gr:
        return pr.PyRanges(df)
    else:
        return df


def run_mosdepth(
    bam_path, 
    t=8, 
    use_median=False, 
    region_bed_path=None, 
    region_gr=None, 
    window_size=None, 
    donot_subset_bam=False, 
    as_gr=True, 
    load_perbase=False,
):
    # sanity check
    if (
        (window_size is not None)
        and (
            (region_bed_path is not None)
            or (region_gr is not None)
        )
    ):  
        raise Exception(f'"window_size" and "region*" arguments must not be used at the same time.')

    if (
        (region_bed_path is not None)
        and (region_gr is not None)
    ):  
        raise Exception(f'Only one of "region_bed_path" and "region_gr" argument must be set.')

    # handle region_gr argument
    region_df = cnvmisc.arg_into_df(region_gr)
    if region_df is not None:
        cnvmisc.genome_df_sanitycheck(region_df)

    # make tmp directory
    tmpdir = tempfile.mkdtemp(dir=os.getcwd(), prefix='mosdepth_tmpdir_')

    # write region bed file
    if (region_bed_path is None) and (region_df is None):
        mosdepth_input_bed_path = None
    else:
        if region_df is not None:
            unedited_input_bed_path = os.path.join(tmpdir, 'unedited_region.bed')
            region_df.to_csv(unedited_input_bed_path, sep='\t', header=False, index=False)
        elif region_bed_path is not None:
            unedited_input_bed_path = region_bed_path

        mosdepth_input_bed_path = os.path.join(tmpdir, 'region.bed')
        with tools.openfile(unedited_input_bed_path, 'r') as infile:
            with open(mosdepth_input_bed_path, 'wt') as outfile:
                for line in infile:
                    linesp = line.strip().split('\t')
                    if not (linesp[1].isdigit() and linesp[2].isdigit()):
                        continue
                    outfile.write('\t'.join(linesp[:3]) + '\n')

    if mosdepth_input_bed_path is None:
        mosdepth_input_bed_gr = None
    else:
        mosdepth_input_bed_gr = pr.PyRanges(pr.read_bed(mosdepth_input_bed_path, as_df=True), int64=False)

    # subset input bam file
    if mosdepth_input_bed_path is not None:
        if donot_subset_bam:
            mosdepth_input_bam_path = bam_path
        else:
            mosdepth_input_bam_path = os.path.join(tmpdir, 'mosdepth_input.bam')
            bameditor.samtools_view(
                in_bam_path=bam_path, 
                out_bam_path=mosdepth_input_bam_path, 
                region_bed_path=mosdepth_input_bed_path, 
                index=True,
            )
    else:
        mosdepth_input_bam_path = bam_path

    # make mosdepth outdir
    outdir = os.path.join(tmpdir, 'outdir')
    os.mkdir(outdir)

    # run mosdepth
    prefix = 'PREFIX'
    mosdepth_args = get_mosdepth_args(
        prefix=os.path.join(outdir, prefix), 
        bam_path=mosdepth_input_bam_path, 
        t=t, 
        use_median=use_median, 
        no_perbase=(not load_perbase),
        bed_path=mosdepth_input_bed_path, 
        window_size=window_size,
    )
    p = subprocess.run(mosdepth_args, capture_output=True, text=True, check=False)
    if p.returncode != 0:
        print(f'stdout: {p.stdout}')
        print(f'stderr: {p.stderr}')
        p.check_returncode()

    # load mosdepth outputs
    outfile_path = os.path.join(outdir, f'{prefix}.regions.bed.gz')
    if use_median:
        depth_colname = 'median_depth'
    else:
        depth_colname = 'mean_depth'
    df = load_mosdepth_output(outfile_path, depth_colname=depth_colname, as_gr=False)

    if load_perbase:
        perbase_outfile_path = os.path.join(outdir, f'{prefix}.per-base.bed.gz')
        df_perbase = pd.read_csv(
            perbase_outfile_path, 
            sep='\t', 
            names=['Chromosome', 'Start', 'End', 'depth'], 
            dtype={'Chromosome': str, 'Start': int, 'End': int, 'depth': int},
        )

        if mosdepth_input_bed_gr is not None:
            df_perbase = mosdepth_input_bed_gr.window(1).join(
                pr.PyRanges(df_perbase, int64=False),
                apply_strand_suffix=False,
            )[['depth']].df
    else:
        df_perbase = None

    # prepare result
    if as_gr:
        result = pr.PyRanges(df, int64=False)
        if load_perbase:
            result_perbase = pr.PyRanges(df_perbase, int64=False)
        else:
            result_perbase = None
    else:
        result = df
        result_perbase = df_perbase

    # remove tmpdir
    shutil.rmtree(tmpdir)

    # return
    if result_perbase is None:
        return result
    else:
        return result, result_perbase





import gzip
import collections
import functools
import operator
import itertools
import re

import pysam

import handygenome.tools as tools
import handygenome.refgenome as refgenome
import handygenome.logutils as logutils
import handygenome.workflow as workflow
import handygenome.hgvs as hgvs
import handygenome.variant.varianthandler as varianthandler
import handygenome.variant.infoformat as infoformat
import handygenome.annotation.annotitem as annotitem
import handygenome.annotation.cosmic as libcosmic
import handygenome.vcfeditor.indexing as indexing
import handygenome.deco as deco


LOGGER = workflow.get_logger(name='COSMIC converter')


def get_header(infile_path):
    with gzip.open(infile_path, "rt") as infile:
        header = tools.get_linesp(next(infile), sep="\t")
    return header


def check_decodable(b):
    try:
        b.decode("utf-8")
    except UnicodeDecodeError:
        return False
    else:
        return True


def get_linedict_byte(byteline, header):
    return dict(zip(header, tools.get_linesp_byte(byteline)))


def get_infile_looper(infile_path):
    def decoding_error_handler(line_bytes, header, NR):
        try:
            linedict_bytes = get_linedict_byte(line_bytes, header)
            if not check_decodable(bytes([linedict_bytes["Mutation CDS"][-1]])):
                linedict_bytes["Mutation CDS"] = linedict_bytes["Mutation CDS"][:-1]
                # cosmic v95 GRCh38 CosmicMutantExport.tsv.gz file contains 
                # a line where 'Mutation CDS' field value contains a non-ascii byte
            linedict_str = {
                key: val.decode("utf-8") for key, val in linedict_bytes.items()
            }
            return linedict_str
        except:
            print("Unexpected decoding error")
            print(f"{NR:,}th line")
            print(line_bytes)
            raise

    header = get_header(infile_path)
    with gzip.open(infile_path, "rb") as infile:
        _ = next(infile)  # skip the header line
        NR = 1
        for line_bytes in infile:
            NR += 1
            try:
                line_str = line_bytes.decode("utf-8")
            except UnicodeDecodeError:
                linedict_str = decoding_error_handler(line_bytes, header, NR)
            else:
                linedict_str = dict(zip(header, tools.get_linesp(line_str)))

            yield linedict_str


def union_all(iter_of_iter):
    return set(itertools.chain.from_iterable(iter_of_iter))


################################
# file type specific functions #
################################

# line_parser
def line_parser_mutantexport(linedict):
    coding_score_key = 'FATHMM score'

    result = dict()

    result['cosv'] = linedict['GENOMIC_MUTATION_ID']
    result['sample_id'] = int(linedict['ID_sample'])
    result['tumor_id'] = int(linedict['ID_tumour'])
    result['site'] = linedict['Primary site']
    result['hgvsg'] = linedict['HGVSG']
    result['hgvsc'] = linedict['HGVSC']
    result['somatic_status'] = linedict['Mutation somatic status']

    if coding_score_key in linedict:
        result['coding_score'] = (None 
                                  if (linedict[coding_score_key] == '') else 
                                  float(linedict[coding_score_key]))
    else:
        result['coding_score'] = None

    result['noncoding_score'] = None

    return result


def line_parser_ncv(linedict):
    coding_score_key = 'FATHMM_MKL_CODING_SCORE'
    noncoding_score_key = 'FATHMM_MKL_NON_CODING_SCORE'

    result = dict()

    result['cosv'] = linedict['GENOMIC_MUTATION_ID']
    result['sample_id'] = int(linedict['ID_SAMPLE'])
    result['tumor_id'] = int(linedict['ID_tumour'])
    result['site'] = linedict['Primary site']
    result['hgvsg'] = linedict['HGVSG']
    result['hgvsc'] = None
    result['somatic_status'] = linedict['Mutation somatic status']

    if coding_score_key in linedict:
        result['coding_score'] = (
                None 
                if (linedict[coding_score_key] == '') else 
                float(linedict[coding_score_key])
                )
    else:
        result['coding_score'] = None

    if noncoding_score_key in linedict:
        result['noncoding_score'] = (
                None 
                if (linedict[noncoding_score_key] == '') else 
                float(linedict[noncoding_score_key])
                )
    else:
        result['noncoding_score'] = None

    return result


################################


def line_filter(parsed_line):
    """exclude if False"""
    if parsed_line['cosv'] == '':
        return False
    elif parsed_line['hgvsg'] == '':
        return False
    else:
        return True


def line_sanity_check(parsed_line):
    pass


def init_cosv_info():
    return {
        #'sampleID': list(),
        'primary_site': dict(),
        'primary_site_somatic': dict(),
        'hgvsg': set(),
        'hgvsc': set(),
        'coding_score': set(),
        'noncoding_score': set(),
    }


def load_datafile(infile_path, line_parser, summary, site_count, logging_lineno=1_000_000):
    """Values of 'Mutation somatic status':
        'Confirmed somatic variant',
        'Not specified',
        'Reported in another cancer sample as somatic',
        'Systematic screen',
        'Variant of unknown origin'
    """
    def update_site_count(site_count, parsed_line, uid):
        site_count.setdefault(parsed_line['site'], set())
        site_count[parsed_line['site']].add(uid)

    def handle_sites(cosv_info, parsed_line, uid):
        cosv_info['primary_site'].setdefault(parsed_line['site'], set())
        cosv_info['primary_site'][parsed_line['site']].add(uid)

        if parsed_line['somatic_status'] == 'Confirmed somatic variant':
            cosv_info['primary_site_somatic'].setdefault(parsed_line['site'], set())
            cosv_info['primary_site_somatic'][parsed_line['site']].add(uid)

    def handle_hgvs(cosv_info, parsed_line):
        cosv_info['hgvsg'].add(parsed_line['hgvsg'])
        if parsed_line['hgvsc'] is not None:
            cosv_info['hgvsc'].add(parsed_line['hgvsc'])

    def handle_scores(cosv_info, parsed_line):
        if parsed_line['coding_score'] is not None:
            cosv_info['coding_score'].add(parsed_line['coding_score'])
        if parsed_line['noncoding_score'] is not None:
            cosv_info['noncoding_score'].add(parsed_line['noncoding_score'])

    # loop over data file
    looper = get_infile_looper(infile_path)
    for linedict in logutils.iter_lineno_logging(looper, LOGGER, logging_lineno):
        line_sanity_check(linedict)
        parsed_line = line_parser(linedict)
        if not line_filter(parsed_line):
            continue

        cosv = parsed_line['cosv']
        uid = (parsed_line['sample_id'], parsed_line['tumor_id'])

        update_site_count(site_count, parsed_line, uid)

        if cosv not in summary:
            summary[cosv] = init_cosv_info()
        cosv_info = summary[cosv]

        handle_sites(cosv_info, parsed_line, uid)
        handle_hgvs(cosv_info, parsed_line)
        handle_scores(cosv_info, parsed_line)

    # postprocess
    #total_count = sum(site_count.values())
    #site_count['total'] = total_count

    #print(f'The number of COSV IDs: {len(summary)}')  # 15689770
    #print(f'The number of COSV IDs with at least one confirmed somatic: {len(summary["primary_sites_somatic"])}')
        # 14577671


###################################################


class MultipleScoreError(Exception):
    pass


def modify_scores(cosv_info, key):
    if len(cosv_info[key]) == 0:
        cosv_info[key] = None
    elif len(cosv_info[key]) == 1:
        cosv_info[key] = cosv_info[key].pop()
    else:
        raise MultipleScoreError(f'More than one values for this cosv entry: {cosv_info}')


def modify_sitedic(site_dic):
    keys = tuple(site_dic.keys())
    for site in keys:
        site_dic[site] = len(site_dic[site])


def modify_vcfspec(cosv_info, fasta, is_hg19):
    def hgvsg_to_vcfspecs(hgvsg_set, fasta):
        vcfspec_candidates = set()
        for hgvsg in hgvsg_set:
            vcfspec = hgvs.hgvsg_to_vcfspec(hgvsg, fasta, leftmost=True)
            #if vcfspec.check_without_N():
            vcfspec_candidates.add(vcfspec)

        return vcfspec_candidates

    def hgvsc_to_vcfspecs(hgvsc_set, fasta, is_hg19):
        vcfspec_candidates = set()
        for hgvsc in hgvsc_set:
            # Running ensembl variantrecoder with hgvsc sometimes results 
                # in error
                # e.g. (v95 GRCh37 CosmicMutantExport.tsv.gz COSV104545830) 
                # Unable to parse HGVS notation 'ENST00000424344.3:c.-166N>A':  
                # : Reference allele extracted from 
                # ENST00000424344:155227447-155227447 (G) does not match 
                # reference allele given by HGVS notation 
                # ENST00000424344.3:c.-166N>A (N)
            try:
                vcfspec = hgvs.hgvsc_to_vcfspec(hgvsc, is_hg19, fasta, leftmost=True)
            except:
                pass
            else:
                #if vcfspec.check_without_N():
                vcfspec_candidates.add(vcfspec)

        return vcfspec_candidates

    def get_vcfspec(cosv_info, fasta, is_hg19):
        """Invalid COSV ID if linked to multiple vcfspecs"""
        vcfspec_candidates = hgvsg_to_vcfspecs(cosv_info['hgvsg'], fasta)
        if len(vcfspec_candidates) > 1:
            vcfspec = None
        elif len(vcfspec_candidates) == 1:
            vcfspec = vcfspec_candidates.pop()
        elif len(vcfspec_candidates) == 0:
            new_vcfspec_candidates = hgvsc_to_vcfspecs(cosv_info['hgvsc'], fasta, is_hg19)
            if len(new_vcfspec_candidates) == 1:
                vcfspec = new_vcfspec_candidates.pop()
            else:
                vcfspec = None

        return vcfspec

    # main
    cosv_info['vcfspec'] = get_vcfspec(cosv_info, fasta, is_hg19)


@deco.get_deco_timestamp('SUMMARY MODIFICATION', LOGGER)
def modify_summary(summary, site_count, fasta, is_hg19, logging_lineno=500_000):
    # summary
    for cosv, cosv_info in logutils.iter_lineno_logging(summary.items(), LOGGER, logging_lineno):
        # coding & noncoding score
        for key in ('coding_score', 'noncoding_score'):
            try:
                modify_scores(cosv_info, key)
            except MultipleScoreError as e:
                raise Exception(f'Error while modifying "{key}"') from e

        # primary site
        modify_sitedic(cosv_info['primary_site'])
        modify_sitedic(cosv_info['primary_site_somatic'])

        # vcfspec
        modify_vcfspec(cosv_info, fasta, is_hg19)

    # site_count
    modify_sitedic(site_count)


###############################################


@deco.get_deco_timestamp('VCFSPEC-TO-COSV MAPPING', LOGGER)
def get_vcfspec_cosv_map(summary, fasta, is_hg19):
    # main
    vcfspec_cosv_map = dict()
    for cosv, cosv_info in summary.items():
        vcfspec = cosv_info['vcfspec']
        if vcfspec is not None:
            vcfspec_cosv_map.setdefault(vcfspec, set())
            vcfspec_cosv_map[vcfspec].add(cosv)

    # deduplicate
    dedup_vcfspec_cosv_map = dict()
    for vcfspec in vcfspec_cosv_map:
        dedup_vcfspec_cosv_map[vcfspec] = vcfspec_cosv_map[vcfspec].pop()

    return dedup_vcfspec_cosv_map


@deco.get_deco_timestamp('VCFSPEC SORTING', LOGGER)
def sort_vcfspecs(vcfspec_cosv_map, chromdict):
    #return sorted(vcfspec_cosv_map.keys(), key=common.get_vcfspec_sortkey(chromdict))
    return sorted(vcfspec_cosv_map.keys(), key=(lambda x: x.get_sortkey(chromdict)))


###############################################


def get_cosmic_metadata(site_count, refver, cosmic_version):
    cosmicmeta = libcosmic.CosmicMetadata()
    cosmicmeta['num_sample_by_site'] = site_count
    cosmicmeta['reference_version'] = refver
    cosmicmeta['cosmic_version'] = cosmic_version

    return cosmicmeta


def get_vcf_header(chromdict, cosmic_metadata):
    header = pysam.VariantHeader()

    for contig, length in chromdict.items():
        if re.fullmatch('(chr)?([0-9]+|X|Y)|chrM|MT', contig) is not None:
            header.contigs.add(contig, length)

    cosmic_metadata.write_header(header)
    libcosmic.CosmicInfoALTlist.add_meta_info(header)
    
    return header


def into_CosmicInfo(cosv, cosv_info):
    cosmicinfo = libcosmic.CosmicInfo()
    cosmicinfo['id'] = cosv
    cosmicinfo['occurrence'] = cosv_info['primary_site']
    cosmicinfo['occurrence_somatic'] = cosv_info['primary_site_somatic']
    cosmicinfo['coding_score'] = cosv_info['coding_score']
    cosmicinfo['noncoding_score'] = cosv_info['noncoding_score']

    cosmicinfolist = libcosmic.CosmicInfoALTlist()
    cosmicinfolist.append(cosmicinfo)

    return cosmicinfolist


@deco.get_deco_timestamp('WRITING FINAL OUTPUT FILE', LOGGER)
def write_outfile(outfile_path, summary, site_count, vcfspec_cosv_map, sorted_vcfspecs, chromdict, refver, cosmic_version, logging_lineno=1_000_000):
    cosmic_metadata = get_cosmic_metadata(site_count, refver, cosmic_version)
    header = get_vcf_header(chromdict, cosmic_metadata)
    with pysam.VariantFile(outfile_path, mode='wz', header=header) as out_vcf:
        for vcfspec in logutils.iter_lineno_logging(sorted_vcfspecs, LOGGER, logging_lineno):
            cosv = vcfspec_cosv_map[vcfspec]
            cosv_info = summary[cosv]
            cosmicinfolist = into_CosmicInfo(cosv, cosv_info)

            vr = header.new_record()
            varianthandler.apply_vcfspec(vr, vcfspec)
            cosmicinfolist.write_info(vr)

            out_vcf.write(vr)

    indexing.index_vcf(outfile_path)


###############################################


@deco.get_deco_arg_choices({'refver': ('GRCh37', 'GRCh38')})
def main(infile_path_ncv, infile_path_mutantexport, outfile_path, refver, cosmic_version):
    fasta = refgenome.get_default_fasta(refver)
    chromdict = refgenome.get_default_chromdict(refver)
    is_hg19 = (refver == 'GRCh37')

    summary = dict()
    site_count = dict()

    LOGGER.info(f'Beginning loading of CosmicMutantExport file')
    load_datafile(infile_path_mutantexport, line_parser_mutantexport, summary, site_count)
    LOGGER.info(f'Finished')

    LOGGER.info(f'Beginning loading of CosmicNCV file')
    load_datafile(infile_path_ncv, line_parser_ncv, summary, site_count)
    LOGGER.info(f'Finished')

    modify_summary(summary, site_count, fasta, is_hg19)
    vcfspec_cosv_map = get_vcfspec_cosv_map(summary, fasta, is_hg19)
    sorted_vcfspecs = sort_vcfspecs(vcfspec_cosv_map, chromdict)

    write_outfile(outfile_path, summary, site_count, vcfspec_cosv_map, sorted_vcfspecs, chromdict, refver, cosmic_version)





import re
import itertools
import logging
import functools

import Bio.Seq

import handygenome.refgenome as refgenome
import handygenome.tools as tools
import handygenome.workflow as workflow
import handygenome.sv.structvars as structvars
import handygenome.variant.vcfspec as libvcfspec


LOGGER = workflow.get_logger(
    name=__name__,
    formatter=logging.Formatter(
        fmt='[%(asctime)s  %(levelname)s] %(name)s - %(message)s',
        datefmt=workflow.DEFAULT_DATEFMT),
    level='info', stderr=True, filenames=None, append=False)


class Breakends:
    """
    Attributes:
        chrom_bnd1
        pos_bnd1
        endis5_bnd1
        chrom_bnd2
        pos_bnd2
        endis5_bnd2
        inserted_seq: 
            - A list composed of the bases of the inserted sequence 
                (e.g. ['A', 'A', 'G', 'C'])
            - As seen in the viewpoint where bnd1-side sequence is on 
                the plus(Crick) strand

        fasta: pysam.FastaFile instance
        chromdict: ChromDict instance
        svtype
        score
        endis3_bnd1
        endis3_bnd2

        _equivs: A list of Breakends objects with maximum reference coverage.
        _pos_range0_bnd1
        _pos_range0_bnd2
        _homlen = None
        _homseq = None
    """

    BASIC_ATTRS = ('chrom_bnd1', 'pos_bnd1', 'endis5_bnd1', 
                   'chrom_bnd2', 'pos_bnd2', 'endis5_bnd2', 
                   'inserted_seq', 'svtype')

    def __init__(self, chrom_bnd1, pos_bnd1, endis5_bnd1, 
                 chrom_bnd2, pos_bnd2, endis5_bnd2, fasta, 
                 inserted_seq=list(), chromdict=None, 
                 svtype=None, score=None):
        # basic attributes
        self.chrom_bnd1 = chrom_bnd1
        self.pos_bnd1 = pos_bnd1
        self.endis5_bnd1 = endis5_bnd1
        self.chrom_bnd2 = chrom_bnd2
        self.pos_bnd2 = pos_bnd2
        self.endis5_bnd2 = endis5_bnd2
        self.inserted_seq = inserted_seq
        self.fasta = fasta

        if chromdict is None:
            self.chromdict = refgenome.ChromDict.from_fasta(fasta)
        else:
            self.chromdict = chromdict

        # endis3
        self.endis3_bnd1 = (not self.endis5_bnd1)
        self.endis3_bnd2 = (not self.endis5_bnd2)

        # svtype
        if svtype is None:
            self._set_svtype()
        else:
            self.svtype = svtype

        # score
        if score is None:
            self.score = 0
        else:
            self.score = score

        # private attributes accesible with getters
        self._equivs = None
        self._homlen = None
        self._homseq = None
        self._pos_range0_bnd1 = None
        self._pos_range0_bnd2 = None

    def __repr__(self):
        def func1(self):
            result = ['<Breakends>']
            result.extend(
                [f'{key}: {getattr(self, key)}' 
                 for key in ('chrom_bnd1', 'pos_bnd1', 'endis5_bnd1', 
                             'inserted_seq',
                             'chrom_bnd2', 'pos_bnd2', 'endis5_bnd2')])
            return '\n'.join(result)

        def func2(self):
            brkt1 = '[' if self.endis5_bnd1 else ']'
            brkt2 = '[' if self.endis5_bnd2 else ']'
            str1 = f'{brkt1}{self.chrom_bnd1}:{self.pos_bnd1}{brkt1}'
            str2 = f'{brkt2}{self.chrom_bnd2}:{self.pos_bnd2}{brkt2}'
            insseq = ''.join(self.inserted_seq)

            return f'<Breakends {str1} {insseq} {str2} (1-based coords)>'

        def func3(self):
            endtype1 = '5\'' if self.endis5_bnd1 else '3\''
            endtype2 = '5\'' if self.endis5_bnd2 else '3\''
            if len(self.inserted_seq) == 0:
                insseq = 'no inserted seq'
            else:
                insseq = ''.join(self.inserted_seq)

            return (f'<Breakends '
                    f'{self.chrom_bnd1}:{self.pos_bnd1} {endtype1} '
                    f'[{insseq}] '
                    f'{self.chrom_bnd2}:{self.pos_bnd2} {endtype2}>')

        return func2(self)

    def __hash__(self):
        return hash((self.chrom_bnd1, self.pos_bnd1, self.endis5_bnd1,
                     self.chrom_bnd2, self.pos_bnd2, self.endis5_bnd2,
                     tuple(self.inserted_seq)))

#    def __hash__(self):
#        bnd1adv_form = get_bnd1adv_form(self)
#        return hash((
#                    bnd1adv_form.chrom_bnd1,
#                    bnd1adv_form.pos_bnd1,
#                    bnd1adv_form.endis5_bnd1,
#                    bnd1adv_form.chrom_bnd2,
#                    bnd1adv_form.pos_bnd2,
#                    bnd1adv_form.endis5_bnd2,
#                    bnd1adv_form.inserted_seq,
#                    ))

    def __eq__(self, other):
        #assert isinstance(other, Breakends)
        return hash(self) == hash(other)

    ###########################################

    def get_params(self, is_bnd1):
        if is_bnd1:
            return (self.chrom_bnd1, 
                    self.get_pos_range0_bnd1(),
                    self.endis5_bnd1)
        else:
            return (self.chrom_bnd2,
                    self.get_pos_range0_bnd2(),
                    self.endis5_bnd2)

    def get_vcfspec_bnd1(self):
        chrom = self.chrom_bnd1
        pos = self.pos_bnd1
        ref = self.fasta.fetch(self.chrom_bnd1, self.pos_bnd1 - 1, 
                               self.pos_bnd1)

        if self.endis5_bnd1:
            t = ''.join(self.inserted_seq) + ref
        else:
            t = ref + ''.join(self.inserted_seq)
        bracket = '[' if self.endis5_bnd2 else ']'
        alt_matestring = f'{bracket}{self.chrom_bnd2}:{self.pos_bnd2}{bracket}'
        if self.endis5_bnd1:
            alt = alt_matestring + t
        else:
            alt = t + alt_matestring

        return libvcfspec.Vcfspec(chrom, pos, ref, [alt])

    def get_vcfspec_bnd2(self):
        chrom = self.chrom_bnd2
        pos = self.pos_bnd2
        ref = self.fasta.fetch(self.chrom_bnd2, self.pos_bnd2 - 1, 
                               self.pos_bnd2)

        insseq_converted = self._convert_seq_between_bnds(
            ''.join(self.inserted_seq))
        if self.endis5_bnd2:
            t = insseq_converted + ref
        else:
            t = ref + insseq_converted

        bracket = '[' if self.endis5_bnd1 else ']'
        alt_matestring = f'{bracket}{self.chrom_bnd1}:{self.pos_bnd1}{bracket}'
        if self.endis5_bnd2:
            alt = alt_matestring + t
        else:
            alt = t + alt_matestring

        return libvcfspec.Vcfspec(chrom, pos, ref, [alt])

    def get_simplesv(self):
        if self.chrom_bnd1 != self.chrom_bnd2:
            raise Exception(f'Translocation cannot be converted to a '
                            f'SimpleStructuralVariant object.')
        else:
            chrom = self.chrom_bnd1
            if self.svtype == 'DEL':
                start1 = self.pos_bnd1 + 1
                end1 = self.pos_bnd2 - 1
                simplesv = structvars.Deletion(chrom=chrom, start1=start1, 
                                               end1=end1, fasta=self.fasta)
            elif self.svtype == 'INV':
                if self.endis5_bnd1:
                    start1 = self.pos_bnd1
                    end1 = self.pos_bnd2 - 1
                else:
                    start1 = self.pos_bnd1 + 1
                    end1 = self.pos_bnd2
                simplesv = structvars.Inversion(chrom=chrom, start1=start1, 
                                                end1=end1, fasta=self.fasta)
            elif self.svtype == 'DUP':
                start1 = self.pos_bnd1
                end1 = self.pos_bnd2
                simplesv = structvars.TandemDuplication(
                    chrom=chrom, start1=start1, end1=end1, fasta=self.fasta)

            return simplesv

    def get_hgvsg(self):
        return self.get_simplesv().get_hgvsg()

    def get_id(self):
        return '_'.join([self.chrom_bnd1,
                         str(self.pos_bnd1),
                         self.chrom_bnd2,
                         str(self.pos_bnd2)])

    def get_id_bnd1(self):
        return self.get_id() + '_1'

    def get_id_bnd2(self):
        return self.get_id() + '_2'

    ###########################################

    @functools.cache
    def get_border_seq(self, is_bnd1):
        (chrom, pos_range0, endis5) = self.get_params(is_bnd1=is_bnd1)
        fetch_start = min(pos_range0)
        fetch_end = max(pos_range0) + 1
        return self.fasta.fetch(chrom, fetch_start, fetch_end)

    @functools.cache
    def get_seq_beyond_bnd_ref(self, is_bnd1, length, with_border_seq):
        (chrom, pos_range0, endis5) = self.get_params(is_bnd1=is_bnd1)
        if endis5:
            fetch_end0 = min(pos_range0)
            fetch_start0 = fetch_end0 - length
        else:
            fetch_start0 = pos_range0.stop
            fetch_end0 = fetch_start0 + length

        seq_beyond = self.fasta.fetch(chrom, fetch_start0, fetch_end0)

        if with_border_seq:
            border_seq = self.get_border_seq(is_bnd1)
            if endis5:
                return seq_beyond + border_seq
            else:
                return border_seq + seq_beyond
        else:
            return seq_beyond

    @functools.cache
    def get_seq_beyond_bnd_alt(self, is_bnd1, length, with_border_seq):
        # get parameters
        endis5 = self.endis5_bnd1 if is_bnd1 else self.endis5_bnd2
        (chrom_mate, pos_range0_mate, endis5_mate) = self.get_params(
            is_bnd1=(not is_bnd1))
        fetch_length = length - len(self.inserted_seq)

        # fetch_start, fetch_end
        if endis5_mate:
            fetch_start = pos_range0_mate.start
            fetch_end = fetch_start + fetch_length
        else:
            fetch_end = pos_range0_mate.start + 1
            fetch_start = fetch_end - fetch_length

        # mateside_seq
        if endis5 == endis5_mate:
            mateside_seq = Bio.Seq.reverse_complement(
                self.fasta.fetch(chrom_mate, fetch_start, fetch_end))
        else:
            mateside_seq = self.fasta.fetch(chrom_mate, fetch_start,
                                            fetch_end)

        # inserted_seq
        if is_bnd1:
            inserted_seq = ''.join(self.inserted_seq)
        else:
            if endis5 == endis5_mate:
                inserted_seq = Bio.Seq.reverse_complement(
                    ''.join(self.inserted_seq))
            else:
                inserted_seq = ''.join(self.inserted_seq)

        # result
        if endis5:
            seq_beyond = mateside_seq + inserted_seq
        else:
            seq_beyond = inserted_seq + mateside_seq

        # add border_seq
        if with_border_seq:
            border_seq = self.get_border_seq(is_bnd1)
            if endis5:
                return seq_beyond + border_seq
            else:
                return border_seq + seq_beyond
        else:
            return seq_beyond

        return result

    ###########################################

    def set_equivs(self):
        self._equivs = get_bnds_equivalents(self)

    def get_equivs(self):
        if self._equivs is None:
            self.set_equivs()
        return self._equivs

    def get_sorted_equivs(self, mode):
        return sort_equivs(self.get_equivs(), mode)

    ###############

    def set_pos_range0s(self):
        poslist0_bnd1 = list()
        poslist0_bnd2 = list()
        for bnds in self.get_equivs():
            poslist0_bnd1.append(bnds.pos_bnd1 - 1)
            poslist0_bnd2.append(bnds.pos_bnd2 - 1)

        if self.endis5_bnd1:
            self._pos_range0_bnd1 = range(max(poslist0_bnd1),
                                          min(poslist0_bnd1) - 1,
                                          -1)
        else:
            self._pos_range0_bnd1 = range(min(poslist0_bnd1),
                                          max(poslist0_bnd1) + 1,
                                          1)

        if self.endis5_bnd2:
            self._pos_range0_bnd2 = range(max(poslist0_bnd2),
                                          min(poslist0_bnd2) - 1,
                                          -1)
        else:
            self._pos_range0_bnd2 = range(min(poslist0_bnd2),
                                          max(poslist0_bnd2) + 1,
                                          1)

    def get_pos_range0_bnd1(self):
        """Returns a directional range"""

        if self._pos_range0_bnd1 is None:
            self.set_pos_range0s()
        return self._pos_range0_bnd1

    def get_pos_range0_bnd2(self):
        """Returns a directional range"""

        if self._pos_range0_bnd2 is None:
            self.set_pos_range0s()
        return self._pos_range0_bnd2

    @functools.cache
    def get_flank_range0(self, mode, is_bnd1, flanklen):
        """Args:
            mode: Must be one of "par", "bnd_prox", or "bnd_dist".
                par: partner side
                bnd_prox: breakend side, proximal (adjacent to the most 
                    advanced border)
                bnd_dist: breakend side, distal (adjacent to the most 
                    retracted border)
        """
        assert mode in ('par', 'bnd_prox', 'bnd_dist')

        # modify flanklen for breakend-side flanks
        if mode in ('bnd_prox', 'bnd_dist'):
            flanklen = flanklen - 1

        # get parameters
        pos_range0 = (self.get_pos_range0_bnd1()
                      if is_bnd1 else
                      self.get_pos_range0_bnd2())

        # main
        step = pos_range0.step

        if mode == 'par':
            start = pos_range0.stop
            stop = start + (step * flanklen)
        else:
            if mode == 'bnd_prox':
                stop = pos_range0.stop - step
            elif mode == 'bnd_dist':
                stop = pos_range0.start

            start = stop + (-step * flanklen)

        return range(start, stop, step)

#    @functools.cache
#    def get_parflank_range0_bnd1(self, flanklen=1):
#        return self._get_flankrange_helper(mode='par', is_bnd1=True,
#                                           flanklen=flanklen)
#
#    @functools.cache
#    def get_parflank_range0_bnd2(self, flanklen=1):
#        return self._get_flankrange_helper(mode='par', is_bnd1=False,
#                                           flanklen=flanklen)
#
#    @functools.cache
#    def get_bndproxflank_range0_bnd1(self, flanklen=1):
#        """Returns a directional range"""
#
#        return self._get_flankrange_helper(mode='bnd_prox', is_bnd1=True,
#                                           flanklen=(flanklen-1))
#
#    @functools.cache
#    def get_bndsideflank_range0_bnd2(self, flanklen=1):
#        """Returns a directional range"""
#
#        return self._get_flankrange_helper(mode=False, is_bnd1=False,
#                                           flanklen=(flanklen-1))

    ###################################################################

    def get_bnd1adv_form(self):
        return pick_bnd1adv_form(self.get_equivs())

    def get_bnd2adv_form(self):
        return pick_bnd2adv_form(self.get_equivs())

    def set_microhomologies(self):
        (self._homlen, 
         self._homseq) = get_microhomology_spec(self.get_equivs())

    def get_homlen(self):
        if self._homlen is None:
            self.set_microhomologies()
        return self._homlen

    def get_homseq(self):
        if self._homseq is None:
            self.set_microhomologies()
        return self._homseq
    
    ###################################################################

    def copy(self):
        return Breakends(
            chrom_bnd1=self.chrom_bnd1, 
            pos_bnd1=self.pos_bnd1, 
            endis5_bnd1=self.endis5_bnd1,
            chrom_bnd2=self.chrom_bnd2, 
            pos_bnd2=self.pos_bnd2, 
            endis5_bnd2=self.endis5_bnd2,
            fasta=self.fasta, chromdict=self.chromdict,
            inserted_seq=self.inserted_seq.copy(), svtype=self.svtype,
            score=self.score)

    def identical(self, other):
        return hash(self) == hash(other)

    def sameseq1(self, other):
        return self.get_bnd1adv_form() == other.get_bnd1adv_form()

    def sameseq2(self, other):
        for key in ('chrom_bnd1', 'endis5_bnd1', 'chrom_bnd2', 'endis5_bnd2'):
            if getattr(self, key) != getattr(other, key):
                return False

        target = self.copy()
        query = other.copy()

        if target.endis5_bnd1:
            goal_bnd1 = max(target.pos_bnd1, query.pos_bnd1)
        else:
            goal_bnd1 = min(target.pos_bnd1, query.pos_bnd1)

        if target.endis5_bnd2:
            goal_bnd2 = max(target.pos_bnd2, query.pos_bnd2)
        else:
            goal_bnd2 = min(target.pos_bnd2, query.pos_bnd2)

        for bnds in (target, query):
            while True:
                if bnds.pos_bnd1 == goal_bnd1:
                    break
                else:
                    bnds.retract_bnd1()
                    continue
            while True:
                if bnds.pos_bnd2 == goal_bnd2:
                    break
                else:
                    bnds.retract_bnd2()
                    continue

        return target.equal(query)

    ###################################################################

    def advance_bnd1(self):
        def get_newpos(self):
            newpos_bnd1 = self._get_advanced_pos_bnd1()
            if len(self.inserted_seq) > 0:
                newpos_bnd2 = self.pos_bnd2
            else:
                newpos_bnd2 = self._get_retracted_pos_bnd2()

            return newpos_bnd1, newpos_bnd2

        def base_match_check(self, newpos_bnd1):
            newbase = self.fasta.fetch(self.chrom_bnd1, newpos_bnd1 - 1, 
                                       newpos_bnd1)
            if len(self.inserted_seq) > 0:
                oldbase = (self.inserted_seq[-1] 
                           if self.endis5_bnd1 else 
                           self.inserted_seq[0])
            else:
                oldbase = self._convert_seq_between_bnds(
                    self.fasta.fetch(self.chrom_bnd2, self.pos_bnd2 - 1, 
                                     self.pos_bnd2))

            return (newbase == oldbase)

        def action(self, newpos_bnd1, newpos_bnd2):
            self.pos_bnd1 = newpos_bnd1
            self.pos_bnd2 = newpos_bnd2
            if len(self.inserted_seq) > 0:
                self.score += 1
                if self.endis5_bnd1:
                    del self.inserted_seq[-1]
                else:
                    del self.inserted_seq[0]

        # main
        newpos_bnd1, newpos_bnd2 = get_newpos(self)
        if self._newpos_range_check(newpos_bnd1, newpos_bnd2):
            if base_match_check(self, newpos_bnd1):
                action(self, newpos_bnd1, newpos_bnd2)
                success = True
            else:
                success = False
        else:
            success = False

        return success

    def advance_bnd2(self):
        def get_newpos(self):
            newpos_bnd2 = self._get_advanced_pos_bnd2()
            if len(self.inserted_seq) > 0:
                newpos_bnd1 = self.pos_bnd1
            else:
                newpos_bnd1 = self._get_retracted_pos_bnd1()

            return newpos_bnd1, newpos_bnd2

        def base_match_check(self, newpos_bnd2):
            newbase = self._convert_seq_between_bnds(
                self.fasta.fetch(self.chrom_bnd2, newpos_bnd2 - 1, 
                                 newpos_bnd2))

            if len(self.inserted_seq) > 0:
                oldbase = (self.inserted_seq[0] 
                           if self.endis5_bnd1 else 
                           self.inserted_seq[-1])
            else:
                oldbase = self.fasta.fetch(self.chrom_bnd1, self.pos_bnd1 - 1, 
                                           self.pos_bnd1)

            return (newbase == oldbase)

        def action(self, newpos_bnd1, newpos_bnd2):
            self.pos_bnd1 = newpos_bnd1
            self.pos_bnd2 = newpos_bnd2
            if len(self.inserted_seq) > 0:
                self.score += 1
                if self.endis5_bnd1:
                    del self.inserted_seq[0]
                else:
                    del self.inserted_seq[-1]

        # main
        newpos_bnd1, newpos_bnd2 = get_newpos(self)
        if self._newpos_range_check(newpos_bnd1, newpos_bnd2):
            if base_match_check(self, newpos_bnd2):
                action(self, newpos_bnd1, newpos_bnd2)
                success = True
            else:
                success = False
        else:
            success = False

        return success

    def retract_bnd1(self):
        added_base = self.fasta.fetch(self.chrom_bnd1, self.pos_bnd1 - 1, 
                                      self.pos_bnd1)
        if self.endis5_bnd1:
            self.inserted_seq.append(added_base)
        else:
            self.inserted_seq.insert(0, added_base)

        self.pos_bnd1 = self._get_retracted_pos_bnd1()
        self.score -= 1

    def retract_bnd2(self):
        added_base = self._convert_seq_between_bnds(
            self.fasta.fetch(self.chrom_bnd2, self.pos_bnd2 - 1, 
                             self.pos_bnd2))

        if self.endis5_bnd1:
            self.inserted_seq.insert(0, added_base)
        else:
            self.inserted_seq.append(added_base)

        self.pos_bnd2 = self._get_retracted_pos_bnd2()
        self.score -= 1

    #######################################################

    def _set_svtype(self):
        if self.chrom_bnd1 != self.chrom_bnd2:
            self.svtype = 'TRA'
        else:
            if (not self.endis5_bnd1) and self.endis5_bnd2:
                self.svtype = 'DEL'
            elif self.endis5_bnd1 and (not self.endis5_bnd2):
                self.svtype = 'DUP'
            elif self.endis5_bnd1 == self.endis5_bnd2:
                self.svtype = 'INV'

    def _newpos_range_check(self, newpos_bnd1, newpos_bnd2):
        return (newpos_bnd1 >= 1 and 
                newpos_bnd1 <= self.chromdict[self.chrom_bnd1] and
                newpos_bnd2 >= 1 and 
                newpos_bnd2 <= self.chromdict[self.chrom_bnd2])

    def _get_advanced_pos_bnd1(self):
        if self.endis5_bnd1:
            return self.pos_bnd1 - 1
        else:
            return self.pos_bnd1 + 1

    def _get_retracted_pos_bnd1(self):
        if self.endis5_bnd1:
            return self.pos_bnd1 + 1
        else:
            return self.pos_bnd1 - 1

    def _get_advanced_pos_bnd2(self):
        if self.endis5_bnd2:
            return self.pos_bnd2 - 1
        else:
            return self.pos_bnd2 + 1

    def _get_retracted_pos_bnd2(self):
        if self.endis5_bnd2:
            return self.pos_bnd2 + 1
        else:
            return self.pos_bnd2 - 1

    def _convert_seq_between_bnds(self, seq):
        return convert_seq_between_bnds(seq, 
                                        self.endis5_bnd1, 
                                        self.endis5_bnd2)


#######################################################


class NonStandardSVAlt(Exception):
    pass


class NonStandardSVRecord(Exception):
    pass


def convert_seq_between_bnds(seq, endis5_bnd1, endis5_bnd2):
    if endis5_bnd1 == endis5_bnd2:
        return Bio.Seq.reverse_complement(seq)
    else:
        return seq


#######################################################


def get_bnds_equivalents(bnds):
    """Returns: A list of Breakends objects, all equivalent to the 
        input object, sorted such that the first item is the most advanced 
        form with respect to bnd1, and the last item is the most advanced 
        form with respect to bnd2.
    """

    input_copy = bnds.copy()

    bnds_list_bnd1adv = list()
    bnds_list_bnd1adv.append(input_copy)
    while True:
        new = bnds_list_bnd1adv[-1].copy()
        success = new.advance_bnd1()
        if success:
            bnds_list_bnd1adv.append(new)
            continue
        else:
            break

    bnds_list_bnd2adv = list()
    bnds_list_bnd2adv.append(input_copy)
    while True:
        new = bnds_list_bnd2adv[-1].copy()
        success = new.advance_bnd2()
        if success:
            bnds_list_bnd2adv.append(new)
            continue
        else:
            break

    bnds_list_bnd1adv.reverse()
    bnds_list = bnds_list_bnd1adv[:-1] + [input_copy] + bnds_list_bnd2adv[1:]
    max_score = max(x.score for x in bnds_list)

    bnds_equivs = [x for x in bnds_list if x.score == max_score]

    return bnds_equivs


def sort_equivs(bnds_equivs, mode):
    assert mode in ('bnd1adv_first', 'bnd1adv_last', 'bnd2adv_first', 
                    'bnd2adv_last'), 'Invalid "mode" argument'

    if mode == 'bnd1adv_first':
        reverse = (not bnds_equivs[0].endis5_bnd1)
        key = lambda bnds: bnds.pos_bnd1
    elif mode == 'bnd1adv_last':
        reverse = bnds_equivs[0].endis5_bnd1
        key = lambda bnds: bnds.pos_bnd1
    elif mode == 'bnd2adv_first':
        reverse = (not bnds_equivs[0].endis5_bnd2)
        key = lambda bnds: bnds.pos_bnd2
    elif mode == 'bnd2adv_last':
        reverse = bnds_equivs[0].endis5_bnd2
        key = lambda bnds: bnds.pos_bnd2

    return sorted(bnds_equivs, key=key, reverse=reverse)


def pick_bnd1adv_form(bnds_equivs):
    if bnds_equivs[0].endis5_bnd1:
        return min(bnds_equivs, key=(lambda bnds: bnds.pos_bnd1))
    else:
        return max(bnds_equivs, key=(lambda bnds: bnds.pos_bnd1))


def pick_bnd2adv_form(bnds_equivs):
    if bnds_equivs[0].endis5_bnd2:
        return min(bnds_equivs, key=(lambda bnds: bnds.pos_bnd2))
    else:
        return max(bnds_equivs, key=(lambda bnds: bnds.pos_bnd2))


def get_microhomology_spec(bnds_equivs):
    homlen = len(bnds_equivs) - 1
    if homlen == 0:
        homseq = ''
    else:
        bnds = bnds_equivs[0]
        if bnds.endis5_bnd1:
            pos_list = [
                x.pos_bnd1 for x in 
                sorted(bnds_equivs, key = lambda x: x.pos_bnd1)[:-1]]
        else:
            pos_list = [
                x.pos_bnd1 for x in 
                sorted(bnds_equivs, key = lambda x: x.pos_bnd1)[1:]]
        homseq = bnds.fasta.fetch(bnds.chrom_bnd1, pos_list[0]-1, 
                                  pos_list[-1])

    return homlen, homseq


#######################################################

def get_bnds_from_vr(vr, fasta, chromdict):
    """
    Raises:
        If input vr does not conform to a known SV variant record format 
            (including a valid non-SV variant record)
    """

    assert len(vr.alts) == 1, (
        f'Multiallelic variant record is not allowed:\n{vr}')

    vr_svinfo = get_vr_svinfo_standard_vr(vr, fasta, chromdict)
    bnds = get_bnds_from_vr_svinfo(vr, vr_svinfo, fasta, chromdict)

    return bnds


########################################################

def get_bnds_from_vr_svinfo(vr, vr_svinfo, fasta, chromdict):
    def warn():
        LOGGER.warning(f'"t" portion of SV ALT string is not an extension '
                       f'of REF string for this variant record:\n{vr}')

    if vr_svinfo['is_bnd1']:
        chrom_bnd1 = vr.contig
        chrom_bnd2 = vr_svinfo['chrom_mate']
        endis5_bnd1 = vr_svinfo['current_endis5']
        endis5_bnd2 = vr_svinfo['mate_endis5']

        if endis5_bnd1:
            if vr_svinfo['t'][-1] == vr_svinfo['ref']:
                pos_bnd1 = vr.pos
                inserted_seq = list( vr_svinfo['t'][:-1] )
            else:
                warn()
                pos_bnd1 = vr.pos + 1
                inserted_seq = list( vr_svinfo['t'] )
        else:
            if vr_svinfo['t'][0] == vr_svinfo['ref']:
                pos_bnd1 = vr.pos
                inserted_seq = list( vr_svinfo['t'][1:] )
            else:
                warn()
                pos_bnd1 = vr.pos - 1
                inserted_seq = list( vr_svinfo['t'] )
            
        pos_bnd2 = vr_svinfo['pos_mate']

    else:
        chrom_bnd1 = vr_svinfo['chrom_mate']
        chrom_bnd2 = vr.contig
        endis5_bnd1 = vr_svinfo['mate_endis5']
        endis5_bnd2 = vr_svinfo['current_endis5']

        pos_bnd1 = vr_svinfo['pos_mate']

        if endis5_bnd2:
            if vr_svinfo['t'][-1] == vr_svinfo['ref']:
                pos_bnd2 = vr.pos
                inserted_seq = list(
                    convert_seq_between_bnds(vr_svinfo['t'][:-1], 
                                             endis5_bnd1, endis5_bnd2))
            else:
                warn()
                pos_bnd2 = vr.pos + 1
                inserted_seq = list(
                    convert_seq_between_bnds(vr_svinfo['t'], 
                                             endis5_bnd1, endis5_bnd2))
        else:
            if vr_svinfo['t'][0] == vr_svinfo['ref']:
                pos_bnd2 = vr.pos
                inserted_seq = list(
                    convert_seq_between_bnds(vr_svinfo['t'][1:], 
                                             endis5_bnd1, endis5_bnd2))
            else:
                warn()
                pos_bnd2 = vr.pos - 1
                inserted_seq = list(
                    convert_seq_between_bnds(vr_svinfo['t'], 
                                             endis5_bnd1, endis5_bnd2))

    bnds = Breakends(chrom_bnd1=chrom_bnd1, pos_bnd1=pos_bnd1,
                     endis5_bnd1=endis5_bnd1,
                     chrom_bnd2=chrom_bnd2, pos_bnd2=pos_bnd2,
                     endis5_bnd2=endis5_bnd2,
                     inserted_seq=inserted_seq, fasta=fasta,
                     chromdict=chromdict)

    return bnds


def get_vr_svinfo_standard_vr(vr, fasta, chromdict):
    vr_svinfo = dict()
    vr_svinfo['ref'] = vr.ref

    try:
        (vr_svinfo['t'],
         vr_svinfo['chrom_mate'],
         vr_svinfo['pos_mate'],
         vr_svinfo['current_endis5'],
         vr_svinfo['mate_endis5']) = parse_sv_altstring(vr.alts[0])
    except NonStandardSVAlt as e:
        e_msg = f'{str(e)}\nInput variant record:\n{vr}'
        raise NonStandardSVRecord(e_msg)

    vr_svinfo['is_bnd1'] = get_is_bnd1(vr, vr_svinfo, chromdict)

    return vr_svinfo


def parse_sv_altstring(sv_altstring):
    mats = [libvcfspec.PAT_BND1.match(sv_altstring), libvcfspec.PAT_BND2.match(sv_altstring)]
    mats_isNotNone = [(x is not None) for x in mats]

    nTrue = mats_isNotNone.count(True)
    if nTrue == 0: # not a bnd string
        raise NonStandardSVAlt(f'ALT string "{sv_altstring}" does not match '
                               f'the standard SV string pattern.') 

    elif nTrue == 1:
        mat = next(itertools.compress(mats, mats_isNotNone))
        t = mat.group('t') # t : according to VCF spec documentation
        chrom_mate = mat.group('matechrom')
        pos_mate = int(mat.group('matepos'))

        if sv_altstring.startswith('[') or sv_altstring.startswith(']'):
            endtype_current_is5 = True
        else:
            endtype_current_is5 = False
        
        if mat.group('bracket1') == '[':
            endtype_mate_is5 = True
        else:
            endtype_mate_is5 = False

        sv_altstring_parsed = (t, chrom_mate, pos_mate, endtype_current_is5, 
                               endtype_mate_is5)

    elif nTrue == 2: # not a valid bnd string
        raise NonStandardSVAlt(f'ALT string "{sv_altstring}" matches both '
                               f'pat1 and pat2.')

    return sv_altstring_parsed


def get_is_bnd1(vr, vr_svinfo, chromdict):
    order = tools.compare_coords(
        vr.contig, vr.pos, vr_svinfo['chrom_mate'], vr_svinfo['pos_mate'], chromdict,
    )
    if order < 0:
        is_bnd1 = True
    elif order > 0:
        is_bnd1 = False
    elif order == 0:
        is_bnd1 = True
#        raise Exception(
#            f'Current chrom/pos and mate chrom/pos are identical '
#            f'for vcf record :\n{vr}')

    return is_bnd1





import os
import pickle
import itertools

import numpy as np
import scipy.stats
import scipy.optimize
import scipy.interpolate
import sklearn.mixture

import handygenome
import handygenome.cnv.misc as cnvmisc


BAFCORRECT_PATH = os.path.join(handygenome.DIRS['data'], f'baf_correction_func.pickle')

def save_bafcorrect_func(func):
    with open(BAFCORRECT_PATH, 'wb') as outfile:
        pickle.dump(func, outfile)


def load_bafcorrect_func(x_cutoff=None):
    with open(BAFCORRECT_PATH, 'rb') as infile:
        func = pickle.load(infile)

    return winsorize_pointfive(func, x_cutoff=x_cutoff)

#

def save_bafcorrect_func_xcutoff(func, x_cutoff):
    with open(BAFCORRECT_PATH, 'wb') as outfile:
        pickle.dump((func, x_cutoff), outfile)

    
def load_bafcorrect_func_xcutoff():
    with open(BAFCORRECT_PATH, 'rb') as infile:
        func, x_cutoff = pickle.load(infile)

    return winsorize_pointfive(func, x_cutoff)


def get_bafs(vaf_list):
    bafs = np.array(vaf_list)
    gt_half_indexes = bafs > 0.5
    bafs[gt_half_indexes] = 1 - bafs[gt_half_indexes]
    return bafs


# baf simulation

def make_vafs_bafs(baf, mean_depth, size):
    halfsize = int(size / 2)
    depths = scipy.stats.poisson.rvs(mu=mean_depth, size=size)
    alts = np.concatenate(
        (
            scipy.stats.binom.rvs(n=depths[:halfsize], p=baf), 
            scipy.stats.binom.rvs(n=depths[halfsize:], p=(1 - baf)), 
        )
    )

    vafs = alts / depths
    np.random.shuffle(vafs)

    bafs = vafs.copy()
    bafs[bafs > 0.5] = 1 - bafs[bafs > 0.5]

    return vafs, bafs


def make_simulated_germline_vafs(mean_depth, size, mode, a=None, b=None):
    depths = scipy.stats.poisson.rvs(mu=mean_depth, size=size)

    if mode == 'betabinom':
        alts = scipy.stats.betabinom.rvs(n=depths, a=a, b=b)
    elif mode == 'binom':
        alts = scipy.stats.binom.rvs(n=depths, p=0.5)

    vafs = alts / depths

    return vafs


def make_simulated_germline_vafs_normal(size, std=0.1):
    result = scipy.stats.norm.rvs(loc=0.5, scale=std, size=size)
    result[result < 0] = 0
    result[result > 1] = 1
    return result


def infer_baf_minimize(baf_values, density=None):
    if density is None:
        density = scipy.stats.gaussian_kde(baf_values)

    argmin = scipy.optimize.minimize_scalar(lambda x: -density(x)[0], bounds=(0, 0.5))
    return argmin.x


def infer_baf_mixture(vaf_values):
    vals = vaf_values.reshape(-1, 1)
    gm = sklearn.mixture.GaussianMixture(n_components=2).fit(vals)
    means = sorted(gm.means_.ravel())
    return (means[0] + (1 - means[1])) / 2


def make_simulation_data(mean_depth=34, size=50, reps=100, true_bafs=None):
    if true_bafs is None:
        true_bafs = np.round(np.arange(0.01, 0.51, 0.01), 2)

    skew_data = dict()
    ebafmean_data = dict()
    ebafmin_data = dict()
    ebafmix_data = dict()
    baf_data = dict()
    vaf_data = dict()

    for baf in true_bafs:
        print(baf)

        skew_values = list()
        ebafmean_values = list()
        ebafmin_values = list()
        ebafmix_values = list()
        vafs_list = list()
        bafs_list = list()

        for _ in range(reps):
            vafs, bafs = make_vafs_bafs(baf=baf, mean_depth=mean_depth, size=size)
            vafs_list.append(vafs)
            bafs_list.append(bafs)
            skew_values.append(scipy.stats.skew(bafs))
            ebafmean_values.append(bafs.mean())
            ebafmin_values.append(infer_baf_minimize(bafs))
            ebafmix_values.append(infer_baf_mixture(vafs))

        skew_data[baf] = np.array(skew_values)
        ebafmean_data[baf] = np.array(ebafmean_values)
        ebafmin_data[baf] = np.array(ebafmin_values)
        ebafmix_data[baf] = np.array(ebafmix_values)
        baf_data[baf] = np.array(bafs_list)
        vaf_data[baf] = np.array(vafs_list)

    return {
        'skew': skew_data,
        'mean': ebafmean_data,
        'minimize': ebafmin_data,
        'mixture': ebafmix_data,
        'baf': baf_data,
        'vaf': vaf_data,
    }


def winsorize_pointfive(func, x_cutoff=None):
    def newfunc(x):
        result = func(x)
        result[result > 0.5] = 0.5
        if x_cutoff is not None:
            result[x > x_cutoff] = 0.5
        return result

    return newfunc


def get_xs_ratio(data, realdata, size=10000, rep=10000):
    realdata_means = np.array([np.random.choice(realdata, size).mean() for x in range(rep)])
    xs_ratio = realdata_means.mean() / data['mean'][0.5].mean()
    return xs_ratio


def get_x_cutoff(realdata, size=100, rep=10000, q=0.05):
    realdata_means = np.array([np.random.choice(realdata, size).mean() for x in range(rep)])
    return np.quantile(realdata_means, q)


def xs_ratio_correction(xs_ratio, xs, ys):
    if xs_ratio is None:
        corrected_xs = None
        interp = scipy.interpolate.interp1d(xs, ys, fill_value='extrapolate')
    else:
        corrected_xs = xs * xs_ratio
        interp = scipy.interpolate.interp1d(corrected_xs, ys, fill_value='extrapolate')

    return interp, corrected_xs


def make_bafcorrector_type1(data, xs_ratio=None):
    xs = list()
    ys = list()
    for true_baf_val, ebaf_vals in data['mean'].items():
        xs.append(ebaf_vals.mean())
        ys.append(true_baf_val)

    xs = np.array(xs)
    ys = np.array(ys)

    interp, corrected_xs = xs_ratio_correction(xs_ratio, xs, ys)
    return interp, xs, ys, corrected_xs


def make_bafcorrector_type2(data, xs_ratio=None):
    xs = list()
    ys = list()
    for true_baf_val, ebaf_vals in data['mean'].items():
        xs.extend(ebaf_vals)
        ys.extend(itertools.repeat(true_baf_val, len(ebaf_vals)))

    xs = np.array(xs)
    ys = np.array(ys)

    interp, corrected_xs = xs_ratio_correction(xs_ratio, xs, ys)

    return interp, xs, ys, corrected_xs


def make_bafcorrector_from_realdata(data, realdata, mode='type1'):
    xs_ratio = get_xs_ratio(data, realdata, size=10000, rep=10000)
    x_cutoff = get_x_cutoff(realdata, size=100, rep=10000, q=0.05)
    if mode == 'type1':
        interp, xs, ys, corrected_xs = make_bafcorrector_type1(data, xs_ratio=xs_ratio)
    elif mode == 'type2':
        interp, xs, ys, corrected_xs = make_bafcorrector_type2(data, xs_ratio=xs_ratio)

    bafcorrector = winsorize_pointfive(interp, x_cutoff)

    return bafcorrector, interp, x_cutoff

        
# baf simulation v2 (230525)

def make_simulated_total_depths(mean_depth, N):
    result = list()
    n = N
    while True:
        arr = scipy.stats.poisson.rvs(mean_depth, size=n)
        arr = arr[arr > 0]
        result.extend(arr)
        if len(result) == N:
            break
        else:
            n = N - len(result)
            continue
    return np.asarray(result)


def make_simulated_het_bafs(vaf, mean_depth, N):
    td = make_simulated_total_depths(mean_depth, N)
    ad = scipy.stats.binom.rvs(n=td, p=vaf)
    vafs = ad / td
    bafs = np.where(vafs > 0.5, 1 - vafs, vafs)
    return bafs


def make_simulated_hom_bafs(p_error, mean_depth, N):
    td = make_simulated_total_depths(mean_depth, N)
    ad = scipy.stats.binom.rvs(n=td, p=p_error)
    vafs = ad / td
    bafs = np.where(vafs > 0.5, 1 - vafs, vafs)
    return bafs


def make_simulated_germline_call_bafs(mean_depth, vaf, N, p_error, hom_portion):
    N_hom = int(N * hom_portion)
    N_het = N - N_hom
    return np.concatenate([
        make_simulated_het_bafs(vaf, mean_depth, N_het),
        make_simulated_hom_bafs(p_error, mean_depth, N_hom),
    ])


def infer_baf_density(bafs, bw, rmzero=True):
    if rmzero:
        bafs = bafs[bafs > 0]
    assert len(bafs) > 0
    if len(bafs) == 1:
        return bafs[0]
    else:
        peak_values, peak_densities, density = cnvmisc.get_density_peaks(bafs, bw_method=bw)
        #return np.average(peak_values, weights=peak_densities)
        if peak_values is None:
            return np.nan
        else:
            return max(peak_values)


def infer_baf_mean(bafs):
    bafs = bafs[bafs > 0]
    return bafs.mean()


def make_simulated_baf_dataset(mean_depth=30, N=10000, p_error=0.001, reps=10, hom_portion=0.6, verbose=False):
    true_bafs = np.arange(0, 0.51, 0.01)
    simbaf_data = list()
    for bafval in true_bafs:
        if verbose:
            print(f'baf value {bafval}')
        subdata = list()
        for _ in range(reps):
            #print(f'repetition {_}')
            simulated_bafs = make_simulated_germline_call_bafs(
                mean_depth, bafval, N, p_error, hom_portion,
            )
            subdata.append(simulated_bafs)
        simbaf_data.append(np.array(subdata))

    simbaf_data = np.stack(simbaf_data, axis=0)

    return simbaf_data


def make_inferred_baf_dataset(simbaf_data, bw=1, verbose=False, infer_method='density'):
    inferred_baf_data = list()
    for truebaf_group in simbaf_data:
        subdata = list()
        for bafvals in truebaf_group:
            if infer_method == 'density':
                inferred_baf = infer_baf_density(bafvals, bw=bw)
            elif infer_method == 'mean':
                inferred_baf = infer_baf_mean(bafvals)
            subdata.append(inferred_baf)
        inferred_baf_data.append(np.array(subdata))
    return np.stack(inferred_baf_data, axis=0)


def find_best_bandwidth(mean_depth):
    true_bafs = np.arange(0, 0.51, 0.01)
    simbaf_data = make_simulated_baf_dataset(
        mean_depth=mean_depth, N=1000, p_error=0.001, reps=10, hom_portion=0.6, verbose=False
    )
    def target(bw):
        inferred_baf_data = make_inferred_baf_dataset(simbaf_data, bw=bw, verbose=False, infer_method='density')
        stdsum = np.std(inferred_baf_data, axis=1).sum()
        means = np.mean(inferred_baf_data, axis=1)
        devsum = np.sqrt(np.sum((true_bafs - means) ** 2))
        return stdsum + devsum

    bwlist = np.arange(0.5, 1, 0.05)
    targetvals = [target(x) for x in bwlist]
    return bwlist[np.argmin(targetvals)]





import os
import re
import contextlib
import pickle
import tempfile

import pysam

import handygenome
import handygenome.tools as tools
import handygenome.network as network
import handygenome.logutils as logutils
#import handygenome.publicdb.ncbi_postprocess as ncbi_postprocess
#import handygenome.fastahandler as fastahandler


URL_AUTHORITY = 'ftp.ncbi.nlm.nih.gov'
URL_PATH_REFSEQ_GENOME = '/genomes/refseq'
URL_PATH_DBSNP = "/snp/latest_release/VCF"
REFSEQ_ACC_PATSTR = r'(?P<refseq_acc>(?P<refseq_acc_prefix>[^_]+)_(?P<refseq_acc_main>[^_]+)\.(?P<refseq_acc_version>[^_]+))'
REFVER_PATSTR = r'(?P<refver>(?P<refver_main>[^.]+)(\.(?P<refver_sub>[^.]+))?)'
GENOME_DIR_PAT = re.compile(REFSEQ_ACC_PATSTR + '_' + REFVER_PATSTR)

#SPECIES_PARDIR_MAP = {
#    'Homo_sapiens': 'vertebrate_mammalian',
#    'Mus_musculus': 'vertebrate_mammalian',
#    'Musa_acuminata': 'plant',
#}


class UnavailableSpeciesError(Exception):
    pass


#################################
# RefSeq ftp genome directories #
#################################

class RefseqPaths:
    species_toppaths_pickle = os.path.join(
        handygenome.DIRS['data'], 'refseq_genome_toppaths.pickle',
    )
    all_genome_supersets = [
        #'assembly_summary_refseq.txt',
        #'assembly_summary_refseq_historical.txt',
        #'README.txt',

        'vertebrate_mammalian',
        'vertebrate_other',
        'plant',
        'bacteria',
        'viral',
        'fungi',
        'invertebrate',
        'protozoa',
        'archaea',

        #'unknown',
        #'metagenomes',
    ]

    def __init__(self):
        self.initialize(force_update=False)

    def remove_species_toppaths_pickle(self):
        os.remove(self.__class__.species_toppaths_pickle)

    def initialize(self, force_update=False):
        if force_update:
            self.remove_species_toppaths_pickle()
        self.init_species_toppaths()
        self.refver_toppaths = dict()
        self.latest_refver_toppaths = dict()
        self.refver_details = dict()

    def init_species_toppaths(self):
        if os.path.exists(self.__class__.species_toppaths_pickle):
            self.load_species_toppaths_pickle()
        else:
            self.species_toppaths = dict()

    #####

    def load_species_toppaths_pickle(self):
        with open(self.__class__.species_toppaths_pickle, 'rb') as f:
            self.species_toppaths = pickle.load(f)

    def save_species_toppaths_pickle(self):
        with open(self.__class__.species_toppaths_pickle, 'wb') as f:
            pickle.dump(self.species_toppaths, f)

    def search_for_species_toppath(self, query_species):
        for superset, subdic in self.species_toppaths.items():
            for species in subdic.keys():
                if species == query_species:
                    return subdic[species]
        return None

    def get_species_toppath(self, species):
        toppath = self.search_for_species_toppath(species)
        if toppath is None:  # now query species is not in the data
            for superset in self.__class__.all_genome_supersets:
                success = self.set_species_toppaths(
                    species_superset=superset,
                    retry_count=None, 
                    retry_interval=5, 
                    timeout=180,  # set a large value for "bacteria"
                )
                toppath = self.search_for_species_toppath(species)
                if toppath is None:
                    continue
                else:
                    break

        if toppath is None:
            raise UnavailableSpeciesError(f'Species "{species}" could not be found from the list of RefSeq genomes')

        return toppath

    def set_species_toppaths(
        self, species_superset,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        if species_superset in self.species_toppaths.keys():
            success = True
        else:
            superset_path = network.join_ftp_paths(URL_PATH_REFSEQ_GENOME, species_superset)
            logutils.print_timestamp(f'Fetching from "{superset_path}"')
            try:
                nlst_result = network.ftp_nlst(
                    URL_AUTHORITY, superset_path,
                    retry_count=retry_count, 
                    retry_interval=retry_interval, 
                    timeout=timeout,
                )
            except network.MaxRetryError as exc:
                logutils.print_timestamp(
                    f'Skipping "{superset_path}" because maximum retry count was exceeded'
                )
                success = False
            else:
                # do update
                self.species_toppaths[species_superset] = dict()
                for species_path in nlst_result:
                    species = species_path.split('/')[-1]
                    self.species_toppaths[species_superset][species] = species_path
                success = True
                # save updated dict to pickle file
                self.save_species_toppaths_pickle()

        return success

    #####

    def set_refver_toppaths(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        species_toppath = self.get_species_toppath(species)
        nlst_path = network.join_ftp_paths(species_toppath, 'all_assembly_versions')

        fname_list = network.ftp_nlst(
            URL_AUTHORITY, nlst_path,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        )

        data = dict()
        for path in fname_list:
            mat = GENOME_DIR_PAT.fullmatch(os.path.basename(path))
            if mat is None:
                continue
            refver = mat.group('refver_main')
            acc_ver = int(mat.group('refseq_acc_version'))
            data.setdefault(refver, dict())
            data[refver][acc_ver] = path

        self.refver_toppaths[species] = data

    def set_latest_refver_toppaths(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        self.set_refver_toppaths(
            species,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        )

        data = dict() 
        for refver, subdic in self.refver_toppaths[species].items():
            data[refver] = max(subdic.items(), key=lambda x: x[0])[1]
        self.latest_refver_toppaths[species] = data

    def get_refver_toppath(
        self, refver, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        """Returns: "/~/~/all_assembly_versions"
        """
        if (species not in self.latest_refver_toppaths):
            self.set_latest_refver_toppaths(
                species, 
                retry_count=retry_count, 
                retry_interval=retry_interval, 
                timeout=timeout,
            )
        if refver not in self.latest_refver_toppaths[species]:
            raise Exception(f'There is no refver named "{refver}" in the species "{species}"')
        return self.latest_refver_toppaths[species][refver]

    def get_latest_refver_toppaths(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        if species not in self.latest_refver_toppaths:
            self.set_latest_refver_toppaths(
                species, 
                retry_count=retry_count, 
                retry_interval=retry_interval, 
                timeout=timeout,
            )
        return self.latest_refver_toppaths[species]

    #####

    def set_refver_details(
        self, refver, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        refver_toppath = self.get_refver_toppath(
            refver, species,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        )
        data = dict()
        for path in network.ftp_nlst(
            URL_AUTHORITY, refver_toppath,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        ):
            if path.endswith('assembly_report.txt'):
                data['assembly_report'] = path
            elif (
                path.endswith('genomic.fna.gz')
                and (not path.endswith('rna_from_genomic.fna.gz'))
            ):
                data['fasta'] = path
            elif path.endswith('genomic.gff.gz'):
                data['gff'] = path
            elif path.endswith('protein.faa.gz'):
                data['protein_fasta'] = path
            elif path.endswith('rm.out.gz'):
                data['repeatmasker'] = path

        self.refver_details[refver] = data

    def get_refver_details(
        self, refver, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        if (refver not in self.refver_details):
            self.set_refver_details(
                refver, species,
                retry_count=retry_count, 
                retry_interval=retry_interval, 
                timeout=timeout,
            )
        return self.refver_details[refver]

    ###############
    # url getters #
    ###############

    def get_url(self, refver, species, key, force_update):
        if force_update:
            self.initialize(force_update=True)
        path = self.get_refver_details(refver, species)[key]
        return f'https://{URL_AUTHORITY}{path}'

    def get_fasta_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'fasta', force_update)

    def get_assembly_report_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'assembly_report', force_update)

    def get_gff_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'gff', force_update)

    def get_repeatmasker_url(self, refver, species, force_update=False):
        return self.get_url(refver, species, 'repeatmasker', force_update)

    #####################################
    # species/refver availability check #
    #####################################

    def check_species_availability(self, species):
        try:
            toppath = self.get_species_toppath(species)
        except UnavailableSpeciesError:
            return False
        else:
            return True

    def get_available_refvers(
        self, species,
        retry_count=None, retry_interval=5, timeout=60,
    ):
        refver_toppaths = self.get_latest_refver_toppaths(
            species,
            retry_count=retry_count, 
            retry_interval=retry_interval, 
            timeout=timeout,
        )
        return tuple(refver_toppaths.keys())


REFSEQPATHS = RefseqPaths()


# convenience functions

#def get_fasta_url(refver, species, force_update=False):
#    return REFSEQPATHS.get_fasta_url(refver, species, force_update)
#
#
#def get_assembly_report_url(refver, species, force_update=False):
#    return REFSEQPATHS.get_assembly_report_url(refver, species, force_update)
#
#
#def get_gff_url(refver, species, force_update=False):
#    return REFSEQPATHS.get_gff_url(refver, species, force_update)


########################################
# RefSeq file downloaders and fetchers #
########################################

def refseqfile_cacher_base(
    topdir, standard_refver, species, path_basename, force_download, get_url_key,
    download_msg=None,
):
    species_subdir = os.path.join(topdir, species)
    os.makedirs(species_subdir, exist_ok=True)
    filepath = os.path.join(species_subdir, path_basename)

    if (not os.path.exists(filepath)) or force_download:
        did_download = True
        if download_msg is None:
            download_msg = (
                f'Downloading {get_url_key} file for reference version "{standard_refver}"'
            )
        logutils.print_timestamp(download_msg)

        url = REFSEQPATHS.get_url(
            refver=standard_refver, 
            species=species, 
            key=get_url_key, 
            force_update=False,
        )
        network.download(url, filepath)
    else:
        did_download = False

    return filepath, did_download


# fasta

FASTAFILE_DIR = os.path.join(handygenome.DIRS['data'], 'refseq_fasta')
os.makedirs(FASTAFILE_DIR, exist_ok=True)

def get_unedited_fasta_path(standard_refver, species, force_download=False):
    filepath, did_download = refseqfile_cacher_base(
        topdir=FASTAFILE_DIR, 
        standard_refver=standard_refver, 
        species=species, 
        path_basename=f'{standard_refver}.unedited.fna.gz', 
        force_download=force_download, 
        get_url_key='fasta',
    )

    return filepath


def get_edited_fasta_path(standard_refver, species, force_download=False):
    edited_fasta_path = os.path.join(FASTAFILE_DIR, f'{standard_refver}.edited.fasta.gz')
    if not os.path.exists(edited_fasta_path):
        unedited_fasta_path = get_unedited_fasta_path(
            standard_refver, species, force_download=force_download,
        )
        from handygenome.publicdb.ncbi_postprocess import postprocess_fasta
        postprocess_fasta(unedited_fasta_path, edited_fasta_path)
    return edited_fasta_path


# assemblyspec

ASSEMBLYFILE_DIR = os.path.join(handygenome.DIRS['data'], 'assembly_reports')
os.makedirs(ASSEMBLYFILE_DIR, exist_ok=True)

def get_assemblyfile_path(standard_refver, species, force_download=False):
    filepath, did_download = refseqfile_cacher_base(
        topdir=ASSEMBLYFILE_DIR, 
        standard_refver=standard_refver, 
        species=species, 
        path_basename=f'{standard_refver}.txt', 
        force_download=force_download, 
        get_url_key='assembly_report',
    )

    return filepath


# repeatmasker

REPEATMASKER_DIR = os.path.join(handygenome.DIRS['data'], 'refseq_repeatmasker')
os.makedirs(REPEATMASKER_DIR, exist_ok=True)

def get_repeatmasker_path(standard_refver, species, force_download=False):
    filepath, did_download = refseqfile_cacher_base(
        topdir=REPEATMASKER_DIR, 
        standard_refver=standard_refver, 
        species=species, 
        path_basename=f'{standard_refver}.rm.out.gz', 
        force_download=force_download, 
        get_url_key='repeatmasker',
    )

    return filepath


# geneset

GENESET_DIR = os.path.join(handygenome.DIRS['data'], 'refseq_geneset')
os.makedirs(GENESET_DIR, exist_ok=True)

def get_geneset_path(standard_refver, species, force_download=False):
    filepath, did_download = refseqfile_cacher_base(
        topdir=GENESET_DIR, 
        standard_refver=standard_refver, 
        species=species, 
        path_basename=f'{standard_refver}.unedited.gff.gz', 
        force_download=force_download, 
        get_url_key='gff',
    )

    return filepath





#########################################


#def genome_path_sorting(
#    species_genome_topdir,
#    retry_count=None, retry_interval=5, timeout=60,
#):
#    fname_list = network.ftp_nlst(
#        URL_AUTHORITY, species_genome_topdir,
#        retry_count=retry_count, 
#        retry_interval=retry_interval, 
#        timeout=timeout,
#    )
#
#    genome_path_dict = dict()
#    for path in fname_list:
#        mat = GENOME_DIR_PAT.fullmatch(os.path.basename(path))
#        if mat is None:
#            continue
#
#        refver = mat.group('refver_main')
#        acc_ver = int(mat.group('refseq_acc_version'))
#        genome_path_dict.setdefault(refver, dict())
#        genome_path_dict[refver][acc_ver] = path
#
#    return genome_path_dict
#
#
#def refver_subver_key(refver_sub):
#    if refver_sub is None:
#        refver_sub_key = -1
#    elif refver_sub.startswith('p'):
#        refver_sub_key = int(refver_sub[1:])
#    else:
#        refver_sub_key = int(refver_sub)
#    return refver_sub_key
#
#
#def pick_latest_genome_path(genome_path_dict):
#    result = dict()
#    for refver, subdic in genome_path_dict.items():
#        result[refver] = max(subdic.items(), key=lambda x: x[0])[1]
#    return result
#
#
#def collect_latest_genome_paths(
#    species_list, 
#    ftp=None, 
#    retry_count=10, 
#    retry_interval=1, 
#    timeout=5,
#):
#    def helper(ftp, species_genome_topdir, result):
#        fname_dict = genome_path_sorting(ftp, species_genome_topdir)
#        latest_genome_paths_part = pick_latest_genome_path(fname_dict)
#        result.update(latest_genome_paths_part)
#
#    # access ftp server and collect file paths
#    result = dict()
#    if ftp is None:
#        ftp = network.ftp_login(
#            URL_AUTHORITY, 
#            retry_count=retry_count, 
#            retry_interval=retry_interval, 
#            timeout=timeout,
#        )
#
#    with contextlib.redirect_stdout('/dev/null'):
#        species_pardir_pairs = [
#            (species, SPECIES_PARDIR_MAP[species])
#            for species in species_list
#        ]
#        for species, pardir in species_pardir_pairs:
#            topdir = URL_PATH_REFSEQ_GENOME + f'/{pardir}/{species}/all_assembly_versions'
#            helper(ftp, topdir, result)
#
#    return result
#
#
#def find_assemblyfile_from_genomepath(ftp, genome_path):
#    fname_list = network.ftp_listdir(ftp, genome_path)
#    results = [x for x in fname_list if x.endswith('assembly_report.txt')]
#    if len(results) != 1:
#        raise Exception(f'Cannot find a unique assembly report file')
#    return results[0]
#
#
#def collect_assemblyfile_urls(retry_count=10, retry_interval=1, timeout=5):
#    ftp = network.ftp_login(
#        URL_AUTHORITY, 
#        retry_count=retry_count, 
#        retry_interval=retry_interval, 
#        timeout=timeout,
#    )
#    latest_genome_paths = collect_latest_genome_paths(ftp=ftp)
#    assemblyfile_paths = {
#        refver: find_assemblyfile_from_genomepath(ftp, genome_path)
#        for refver, genome_path in latest_genome_paths.items()
#    }
#
#    # add prefix
#    assemblyfile_urls = {
#        key: f'https://{URL_AUTHORITY}' + val
#        for key, val in assemblyfile_paths.items()
#    }
#
#    # quit ftp
#    ftp.quit()
#
#    return assemblyfile_urls


def get_dbsnp_urls(retry_count=10, retry_interval=1, timeout=5):
    # access ftp server and collect file paths
    ftp = network.ftp_login(
        URL_AUTHORITY, 
        retry_count=retry_count, 
        retry_interval=retry_interval, 
        timeout=timeout,
    )
    with contextlib.redirect_stdout('/dev/null'):
        fname_list = network.ftp_listdir(ftp, URL_PATH_DBSNP)
        ftp.quit()

    # make result
    pat = re.compile(r'(.+)\.gz(.*)')
    url_dict = dict()
    for fname in fname_list:
        mat = pat.fullmatch(os.path.basename(fname))
        if mat is None:  # CHECKSUM
            continue

        key = mat.group(1)
        url_dict.setdefault(key, dict())

        if mat.group(2) == '':
            subkey = 'vcf'
        elif mat.group(2) == '.md5':
            subkey = 'vcf_md5'
        else:
            continue

        url_dict[key][subkey] = f'https://{URL_AUTHORITY}' + fname

    return url_dict




import re

svALT_pat_list = [
	re.compile('^(.+)\[(.+):(.+)\[$'),
	re.compile('^(.+)\](.+):(.+)\]$'),
	re.compile('^\[(.+):(.+)\[(.+)$'),
	re.compile('^\](.+):(.+)\](.+)$'),
	]


def get_SVTYPE(vp):
	if vp.INFO_has_field('SVTYPE'):
		return vp.variant.INFO['SVTYPE']


def get_svALT(vp):


def main(vp):
	vp.SVTYPE = get_SVTYPE(vp)



import os
import gzip

import pysam
import numpy as np

import handygenome.refgenome as refgenome
import handygenome.interval as libinterval


# pysam VariantFile mode related ones

"""
pysam file mode string

- Valid mode string pattern : ^[rwa]([bzu]?[0-9]?|[0-9]?[bzu]?)$
- Conflicting characters: (b,z,u)

- 'wb' : Compressed BCF (level 6), regardless of file name. 
- 'wb[0-9]' : BCF with level of compression indicated by the number, regardless of file name. 

- 'wz' : Compressed VCF (level 6), regardless of file name.
- 'wz[0-9]' : VCF with level of compression indicated by the number, regardless of file name. 

- 'wu' : Uncompressed VCF, regardless of file name.
- 'wu[0-9]' : Uncompressed VCF, regardless of file name.

- 'w[0-9]' : Uncompressed VCF, regardless of file name. 

- 'w' :
    *.vcf : uncompressed VCF
    *.vcf.gz : compressed VCF (level 6)
    *.bcf : compressed BCF (level 6)
    *.bcf.gz : compressed VCF (level 6)
"""

CYVCF2_FORMAT_DICT = {'v':'w', 'z':'wz', 'u':'wbu', 'b':'wb'}  
    # keys: bcftools format strings
    # values: cyvcf2 format strings
PYSAM_MODE_DICT = {'v':'wz0', 'z':'wz', 'u':'wb0', 'b':'wb'}
    # keys: bcftools format strings
    # values: pysam format strings
#DEFAULT_MODE_BCFTOOLS = 'z'


def write_mode_arghandler(mode_bcftools, mode_pysam):
    """
    Args:
        mode_bcftools: bcftools style mode string
        mode_pysam: pysam style mode string

    If 'mode_pysam' is set as a non-None value,
    it overrides 'mode_bcftools'.
    """
    if mode_pysam is None:
        return PYSAM_MODE_DICT[mode_bcftools]
    else:
        return mode_pysam

###

def get_vcf_format(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        is_vcf = vcf.is_vcf
        #is_bcf = vcf.is_bcf
        comp = vcf.compression

    if comp not in ('NONE', 'GZIP', 'BGZF'):
        raise Exception(f'Unexpected compression type')

    is_bgzf = (comp == 'BGZF')

    return is_vcf, comp, is_bgzf


def check_has_index(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        has_index = (vcf.index is not None)
    return has_index


def get_indexfile_path(vcf_path):
    if os.path.exists(f'{vcf_path}.csi'):
        return f'{vcf_path}.csi'
    elif os.path.exists(f'{vcf_path}.tbi'):
        return f'{vcf_path}.tbi'
    else:
        return None


def make_index(vcf_path):
    index_filename = f'{vcf_path}.csi'
    pysam.tabix_index(vcf_path, preset='vcf', csi=True, index=index_filename)
    return index_filename


def get_vcf_positions(vcf_path, as_iter=False, verbose=False):
    """Returns:
        as_iter==True: A list of tuples (chrom, start0, end0) of each variant record
        as_iter==False: np.ndarray
    """
    is_vcf, comp, is_bgzf = get_vcf_format(vcf_path)
    if is_vcf:
        line_iter = get_line_iter(vcf_path, comp, verbose)
        result_iter = parse_line_iter(line_iter)
    else:  # BCF
        result_iter = get_vcf_positions_pysam(vcf_path)

    if as_iter:
        return result_iter
    else:
        #return list(result_iter)
        return np.array(
            list(result_iter), 
            dtype=[('Chromosome', object), ('Start', int), ('End', int)],
        )


def split_vcf_positions(vcf_positions, n):
    """Splits such that adjacent sub-positions do not meet each other with
    identical POS values
    """
    diffresult = np.diff(vcf_positions['Start'])
    nondiff_indexes = np.where(diffresult == 0)[0] + 1
    spittable_indexes = np.setdiff1d(np.arange(vcf_positions.shape[0]), nondiff_indexes)
    spittable_indexes_split = [
        x for x in np.array_split(spittable_indexes, n) 
        if x.shape[0] != 0
    ]
    splitpoints = [x[0] for x in spittable_indexes_split if x[0] != 0]
    split_positions = np.split(vcf_positions, splitpoints)
        # this works even when "splitpoints" is an empty list

    return split_positions


def get_fetchargs_from_vcf_positions(vcf_positions, refver):
    chrom_left = vcf_positions[0]['Chromosome']
    start0_left = vcf_positions[0]['Start']
    chrom_right = vcf_positions[-1]['Chromosome']
    end0_right = vcf_positions[-1]['Start'] + 1  
        # only the 1-base position indicated by POS value is used
    chromdict = refgenome.get_default_chromdict(refver)

    intvlist = libinterval.IntervalList.from_margin(
        chromdict, chrom_left, start0_left, chrom_right, end0_right,
    )
    fetchargs_list = [
        (intv.chrom, intv.start0, intv.end0) for intv in intvlist
    ]
    return fetchargs_list


def get_vcf_fetchregions(vcf_path, n, refver, verbose=False):
    all_position_info = get_vcf_positions(vcf_path, as_iter=False, verbose=verbose)
    split_position_info = split_vcf_positions(all_position_info, n)
    result = [
        get_fetchargs_from_vcf_positions(position_info, refver)
        for position_info in split_position_info
    ]

    return result


def get_vcf_lineno(vcf_path, verbose=False):
    is_vcf, comp, is_bgzf = get_vcf_format(vcf_path)
    if is_vcf:
        line_iter = get_line_iter(vcf_path, comp, verbose)
        return count_iterator(line_iter)
    else:  # BCF
        return get_vcf_lineno_pysam(vcf_path)


def get_vr_fetcher(vcf, refver, chrom=None, start0=None, end0=None, respect_refregion=False):
    if chrom is None:
        for vr in vcf.fetch():
            yield vr
    else:
        if start0 is None:
            start0 = 0
        if end0 is None:
            end0 = refgenome.get_default_chromdict(refver)[chrom]
        
        fetcher = vcf.fetch(contig=chrom, start=start0, stop=end0)

        if respect_refregion:
            def check_vr_inclusion(vr, chrom, start0, end0):
                vr_start0 = vr.pos - 1
                vr_end0 = vr_start0 + len(vr.ref)
                return (
                    (vr.contig == chrom)
                    and (vr_end0 > start0)
                    and (vr_start0 < end0)
                )
        else:
            def check_vr_inclusion(vr, chrom, start0, end0):
                vr_start0 = vr.pos - 1
                vr_end0 = vr.pos
                return (
                    (vr.contig == chrom)
                    and (vr_end0 > start0)
                    and (vr_start0 < end0)
                )

        for vr in fetcher:
            if check_vr_inclusion(vr, chrom, start0, end0):
                yield vr

###

def count_iterator(iterator):
    idx = -1
    for idx, x in enumerate(iterator):
        pass
    return idx + 1


def get_line_iter(vcf_path, comp, verbose):
    if comp in ('BGZF', 'GZIP'):
        raw_iter = gzip.open(vcf_path, 'rb')
    elif comp == 'NONE':
        raw_iter = open(vcf_path, 'rb')

    if verbose:
        for idx, line in enumerate(raw_iter):
            if idx % 100000 == 0:
                print(idx)
            if not line.startswith(b'#'):
                yield line

    else:
        for line in raw_iter:
            if not line.startswith(b'#'):
                yield line


def parse_line_iter(line_iter):
    """Accepts byte mode iterator"""
    for line in line_iter:
        linesp = line.split(b'\t')
        pos0 = int(linesp[1]) - 1
        yield (linesp[0].decode(), pos0, pos0 + len(linesp[3]))


def get_vcf_positions_pysam(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        for vr in vcf.fetch():
            yield (vr.contig, vr.pos - 1, vr.pos - 1 + len(vr.ref))


def get_vcf_lineno_pysam(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        lineno = count_iterator(vcf.fetch())
    return lineno


# this does not work
def get_vcf_noerr(*args, **kwargs):
    with contextlib.redirect_stderr(io.StringIO()) as err, \
            contextlib.redirect_stdout(io.StringIO()) as out:
        vcf = pysam.VariantFile(*args, **kwargs)

    for buf in (err, out):
        msg = buf.getvalue()
        if not msg.startswith('[E::idx_find_and_load] '
                              'Could not retrieve index file for'):
            print(msg, end='', file=sys.stderr)

    return vcf





import functools
import time
import inspect
import itertools
import signal

import numpy as np


# timeout (https://daeguowl.tistory.com/139)
def get_deco_timeout(seconds, error_message=''):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)
        
        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            #signal.alarm(seconds)
            signal.setitimer(signal.ITIMER_REAL, seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return functools.wraps(func)(wrapper)

    return decorator


def deco_timer(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time
        print(run_time)

        return value

    return wrapper_timer


def get_deco_num_set(names, n):
    def decorator(func):
        sig = inspect.signature(func)
        if not set(names).issubset(sig.parameters.keys()):
            raise Exception(
                f'The names of parameters given to '
                f'"get_deco_num_set" function '
                f'is not included in the parameter names of '
                f'the function "{func.__name__}".')

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            n_set = sum((name in ba.arguments) for name in names)
            if n_set != n:
                raise ValueError(
                    f'For the function "{func.__name__}", the '
                    #f'number of parameters set from arguments '
                    f'number of parameters being set, '
                    f'among {tuple(names)}, must be {n}.')

            return func(*args, **kwargs)

        return wrapper

    return decorator


def get_deco_num_notNone(names, n):
    def decorator(func):
        sig = inspect.signature(func)
        if not set(names).issubset(sig.parameters.keys()):
            raise Exception(
                f'The names of parameters given to '
                f'"get_deco_num_set_differently" function '
                f'is not included in the parameter names of '
                f'the function "{func.__name__}".')

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()

            n_notNone = 0
            for name in names:
                set_val = ba.arguments[name]
                if set_val is not None:
                    n_notNone += 1

            if n_notNone != n:
                raise ValueError(
                    f'For the function "{func.__name__}", the '
                    f'number of parameters, among {tuple(names)}, '
                    f'being set as a value different from the default, '
                    f'must be {n}.')

            return func(*args, **kwargs)

        return wrapper

    return decorator


def get_deco_num_set_differently(names, n, how='equal'):
    assert how in ('equal', 'gt', 'lt')
    def decorator(func):
        sig = inspect.signature(func)
        if not set(names).issubset(sig.parameters.keys()):
            raise Exception(
                f'The names of parameters given to '
                f'"get_deco_num_set_differently" function '
                f'is not included in the parameter names of '
                f'the function "{func.__name__}".')

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()

            n_diff = 0
            for name in names:
                default_val = sig.parameters[name].default
                set_val = ba.arguments[name]
                if default_val is None:
                    if set_val is not None:
                        n_diff += 1
                else:
                    if set_val != default_val:
                        n_diff += 1

            if how == 'equal':
                cond = (n_diff == n)
                word = 'equal to'
            elif how == 'gt':
                cond = (n_diff > n)
                word = 'greater than'
            elif how == 'lt':
                cond = (n_diff < n)
                word = 'less than'

            if not cond:
                raise ValueError(
                    f'For the function "{func.__name__}", the '
                    f'number of parameters, among {tuple(names)}, '
                    f'being set as a value different from the default, '
                    f'must be {word} {n}.')

            return func(*args, **kwargs)

        return wrapper

    return decorator


def get_deco_arg_choices(mapping):
    """Args:
        mapping: {'argname': (valid_value1, valid_value2, ...), ...}
    """

    def decorator(func):
        sig = inspect.signature(func)
        if not set(mapping.keys()).issubset(sig.parameters.keys()):
            raise Exception(
                f'The names of parameters given to '
                f'"get_deco_check_arg_choices" function '
                f'is not included in the parameter names of '
                f'the function "{func.__name__}".')

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()
            for key, val in mapping.items():
                if ba.arguments[key] not in val:
                    raise ValueError(
                        f'For the function "{func.__name__}", '
                        f'the parameter "{key}" must be one of these values: '
                        f'{tuple(val)}.')

            return func(*args, **kwargs)

        return wrapper

    return decorator

# other decorators
def get_deco_timestamp(msg, logger):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            logger.info(f'BEGINNING {msg}')
            result = func(*args, **kwargs)
            logger.info(f'FINISHED {msg}')
            return result
        return wrapper

    return decorator


def scalarize_result(result):
    if isinstance(result, (list, tuple)):
        result = tuple(
            x[0] if x.shape == (1,) else x
            for x in result
        )
    else:
        if result.shape == (1,):
            result = result[0]


def vectorize(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        keys = list(kwargs.keys())
        arglist = list(itertools.chain(args, kwargs.values()))
        arglist = np.broadcast_arrays(arglist)

        new_args = arglist[:len(args)]
        new_kwargs = dict(zip(keys, arglist[len(args):]))

        result = func(*new_args, **new_kwargs)
        if isinstance(result, (list, tuple)):
            result = tuple(
                x[0] if x.shape == (1,) else x
                for x in result
            )
        else:
            if result.shape == (1,):
                result = result[0]

        return result

    return wrapper


def args_into_array(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        keys = list(kwargs.keys())
        arglist = [
            np.atleast_1d(x) for x in 
            itertools.chain(args, kwargs.values())
        ]

        new_args = arglist[:len(args)]
        new_kwargs = dict(zip(keys, arglist[len(args):]))

        result = np.squeeze(func(*new_args, **new_kwargs))

        return result

    return wrapper


def get_deco_atleast1d(names):
    def decorator(func):
        sig = inspect.signature(func)
        if not set(names).issubset(sig.parameters.keys()):
            raise Exception(
                f'The names of parameters given to '
                f'"get_deco_atleast1d" function '
                f'is not included in the parameter names of '
                f'the function "{func.__name__}".')

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()
            argdict = ba.arguments
            for key in names:
                argdict[key] = np.atleast_1d(argdict[key])

            return func(**argdict)

        return wrapper

    return decorator


def get_deco_broadcast(names):
    def decorator(func):
        sig = inspect.signature(func)
        if not set(names).issubset(sig.parameters.keys()):
            raise Exception(
                f'The names of parameters given to '
                f'"get_deco_broadcast" function '
                f'is not included in the parameter names of '
                f'the function "{func.__name__}".')

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()
            argdict = ba.arguments

            bc_args = np.broadcast_arrays(*[argdict[key] for key in names])
            for key, newarg in zip(names, bc_args):
                argdict[key] = newarg

            return func(**argdict)

        return wrapper

    return decorator


def get_deco_asarray(names):
    def decorator(func):
        sig = inspect.signature(func)
        if not set(names).issubset(sig.parameters.keys()):
            raise Exception(
                f'The names of parameters given to '
                f'"get_deco_asarray" function '
                f'is not included in the parameter names of '
                f'the function "{func.__name__}".')

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()
            argdict = ba.arguments
            for key in names:
                argdict[key] = np.asarray(argdict[key])

            return func(**argdict)

        return wrapper

    return decorator
